from __future__ import absolute_import, division, print_function

import os

from vivarium.core.process import Deriver


class AgentNames(Deriver):

    name = 'agent_names'
    defaults = {}

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}
        super(AgentNames, self).__init__(initial_parameters)

    def ports_schema(self):
        return {
            'agents': {
                '*': {}
            },
            'names': {
                '_default': [],
                '_updater': 'set',
                '_emit': True,
            },
        }

    def next_update(self, timestep, states):
        agents = states['agents']
        return {'names': list(agents.keys())}



"""
======================================
Toy Glucose Phosphorylation Experiment
======================================

This is a toy example referred to by the documentation.
"""

from vivarium.core.experiment import Experiment
from vivarium_cell.composites.injected_glc_phosphorylation import (
    InjectedGlcPhosphorylation,
)

def glucose_phosphorylation_experiment(config=None):
    if config is None:
        config = {}
    default_config = {
        'injected_glc_phosphorylation': {},
        'emitter': {
            'type': 'timeseries',
        },
        'initial_state': {},
    }
    default_config.update(config)
    config = default_config
    compartment = InjectedGlcPhosphorylation(
        config['injected_glc_phosphorylation'])
    compartment_dict = compartment.generate()
    experiment = Experiment({
        'processes': compartment_dict['processes'],
        'topology': compartment_dict['topology'],
        'emitter': config['emitter'],
        'initial_state': config['initial_state'],
    })
    return experiment



from __future__ import absolute_import, division, print_function

import os

# Python imports
import random
import math
import numpy as nppython

# pymunk imports
import pymunkoptions
pymunkoptions.options["debug"] = False
import pymunk


PI = math.pi

DEBUG_SIZE = 600  # size of the pygame debug screen

def get_force_with_angle(force, angle):
    x = force * math.cos(angle)
    y = force * math.sin(angle)
    return [x, y]


def front_from_corner(width, length, corner_position, angle):
    half_width = width/2
    dx = length * math.cos(angle) + half_width * math.cos(angle + PI/2)  # PI/2 gives a half-rotation for the width component
    dy = length * math.sin(angle) + half_width * math.sin(angle + PI/2)
    front_position = [corner_position[0] + dx, corner_position[1] + dy]
    return np.array([front_position[0], front_position[1], angle])


def corner_from_center(width, length, center_position, angle):
    half_length = length/2
    half_width = width/2
    dx = half_length * math.cos(angle) + half_width * math.cos(angle + PI/2)
    dy = half_length * math.sin(angle) + half_width * math.sin(angle + PI/2)
    corner_position = [center_position[0] - dx, center_position[1] - dy]
    return np.array([corner_position[0], corner_position[1], angle])


def random_body_position(body):
    ''' pick a random point along the boundary'''
    width, length = body.dimensions
    if random.randint(0, 1) == 0:
        # force along ends
        if random.randint(0, 1) == 0:
            # force on the left end
            location = (random.uniform(0, width), 0)
        else:
            # force on the right end
            location = (random.uniform(0, width), length)
    else:
        # force along length
        if random.randint(0, 1) == 0:
            # force on the bottom end
            location = (0, random.uniform(0, length))
        else:
            # force on the top end
            location = (width, random.uniform(0, length))
    return location


class NullScreen(object):
    def update_screen(self):
        pass
    def configure(self, config):
        pass


class PymunkMultibody(object):
    """
    Multibody object for interfacing with pymunk
    """

    defaults = {
        'agent_shape': 'segment',
        # hardcoded parameters
        'elasticity': 0.9,
        'damping': 0.5,  # 1 is no damping, 0 is full damping
        'angular_damping': 0.8,
        'friction': 0.9,  # does this do anything?
        'physics_dt': 0.001,
        'force_scaling': 5e4,  # scales from pN
        # configured parameters
        'jitter_force': 1e-3,  # pN
        'bounds': [20, 20],
        'barriers': False,
        'initial_agents': {},
        # for debugging
        'screen': None,
    }

    def __init__(self, config):
        # hardcoded parameters
        self.elasticity = self.defaults['elasticity']
        self.friction = self.defaults['friction']
        self.damping = self.defaults['damping']
        self.angular_damping = self.defaults['angular_damping']
        self.physics_dt = config.get('physics_dt', self.defaults['physics_dt'])
        self.force_scaling = self.defaults['force_scaling']

        # configured parameters
        self.agent_shape = config.get('agent_shape', self.defaults['agent_shape'])
        self.jitter_force = config.get('jitter_force', self.defaults['jitter_force'])
        self.bounds = config.get('bounds', self.defaults['bounds'])
        barriers = config.get('barriers', self.defaults['barriers'])

        # initialize pymunk space
        self.space = pymunk.Space()

        # debug screen
        self.screen = config.get('screen')
        if self.screen is None:
            self.screen = NullScreen()
        self.screen.configure({
            'space': self.space,
            'bounds': self.bounds})

        # add static barriers
        self.add_barriers(self.bounds, barriers)

        # initialize agents
        initial_agents = config.get('initial_agents', self.defaults['initial_agents'])
        self.bodies = {}
        for agent_id, specs in initial_agents.items():
            self.add_body_from_center(agent_id, specs)

    def run(self, timestep):
        if self.physics_dt > timestep:
            print('timestep skipped by pymunk_multibody: {}'.format(timestep))
            return

        time = 0
        while time < timestep:
            time += self.physics_dt

            # apply forces
            for body in self.space.bodies:
                self.apply_jitter_force(body)
                self.apply_motile_force(body)
                self.apply_viscous_force(body)

            # run for a physics timestep
            self.space.step(self.physics_dt)

        self.screen.update_screen()

    def apply_motile_force(self, body):
        width, length = body.dimensions
        motile_location = (width / 2, 0)  # apply force at back end of body
        thrust = 0.0
        torque = 0.0
        motile_force = [thrust, torque]

        if hasattr(body, 'thrust'):
            thrust = body.thrust
            torque = body.torque
            motile_force = [thrust, 0.0]

            # add to angular velocity
            body.angular_velocity += torque

        scaled_motile_force = [force * self.force_scaling for force in motile_force]
        body.apply_impulse_at_local_point(scaled_motile_force, motile_location)

    def apply_jitter_force(self, body):
        jitter_location = random_body_position(body)
        jitter_force = [
            random.normalvariate(0, self.jitter_force),
            random.normalvariate(0, self.jitter_force)]
        scaled_jitter_force = [
            force * self.force_scaling
            for force in jitter_force]
        body.apply_impulse_at_local_point(
            scaled_jitter_force,
            jitter_location)

    def apply_viscous_force(self, body):
        # dampen velocity
        body.velocity = body.velocity * self.damping + (body.force / body.mass) * self.physics_dt

        # dampen angular velocity
        body.angular_velocity = body.angular_velocity * self.angular_damping + (body.torque / body.moment) * self.physics_dt

    def add_barriers(self, bounds, barriers):
        """ Create static barriers """
        thickness = 5.0
        offset = thickness
        x_bound = bounds[0]
        y_bound = bounds[1]

        static_body = self.space.static_body
        static_lines = [
            pymunk.Segment(
                static_body,
                (0.0-offset, 0.0-offset),
                (x_bound+offset, 0.0-offset),
                thickness),
            pymunk.Segment(
                static_body,
                (x_bound+offset, 0.0-offset),
                (x_bound+offset, y_bound+offset),
                thickness),
            pymunk.Segment(
                static_body,
                (x_bound+offset, y_bound+offset),
                (0.0-offset, y_bound+offset),
                thickness),
            pymunk.Segment(
                static_body,
                (0.0-offset, y_bound+offset),
                (0.0-offset, 0.0-offset),
                thickness),
        ]

        if barriers:
            spacer_thickness = barriers.get('spacer_thickness', 0.1)
            channel_height = barriers.get('channel_height')
            channel_space = barriers.get('channel_space')
            n_lines = math.floor(x_bound/channel_space)

            machine_lines = [
                pymunk.Segment(
                    static_body,
                    (channel_space * line, 0),
                    (channel_space * line, channel_height), spacer_thickness)
                for line in range(n_lines)]
            static_lines += machine_lines

        for line in static_lines:
            line.elasticity = 0.0  # bounce
            line.friction = 0.8
        self.space.add(static_lines)

    def get_shape(self, boundary):
        '''
        shape documentation at: https://pymunk-tutorial.readthedocs.io/en/latest/shape/shape.html
        '''

        if self.agent_shape == 'segment':
            width = boundary['width']
            length = boundary['length']

            half_width = width / 2
            half_length = length / 2 - half_width
            shape = pymunk.Segment(
                None,
                (-half_length, 0),
                (half_length, 0),
                radius=half_width)

        elif self.agent_shape == 'circle':
            length = boundary['length']
            half_length = length / 2
            shape = pymunk.Circle(None, radius=half_length, offset=(0, 0))

        elif self.agent_shape == 'rectangle':
            width = boundary['width']
            length = boundary['length']
            half_length = length / 2
            half_width = width / 2
            shape = pymunk.Poly(None,
                ((-half_length, -half_width),
                 (half_length, -half_width),
                 (half_length, half_width),
                 (-half_length, half_width)))

        return shape

    def get_inertia(self, shape, mass):
        if self.agent_shape == 'rectangle':
            inertia = pymunk.moment_for_poly(mass, shape.get_vertices())
        elif self.agent_shape == 'circle':
            radius = shape.radius
            inertia = pymunk.moment_for_circle(mass, radius, radius)
        elif self.agent_shape == 'segment':
            a = shape.a
            b = shape.b
            radius = shape.radius
            inertia = pymunk.moment_for_segment(mass, a, b, radius)

        return inertia

    def add_body_from_center(self, body_id, specs):
        boundary = specs['boundary']
        mass = boundary['mass']
        center_position = boundary['location']
        angle = boundary['angle']
        angular_velocity = boundary.get('angular_velocity', 0.0)
        width = boundary['width']
        length = boundary['length']

        # get shape, inertia, make body, assign body to shape
        shape = self.get_shape(boundary)
        inertia = self.get_inertia(shape, mass)
        body = pymunk.Body(mass, inertia)
        shape.body = body

        body.position = (
            center_position[0],
            center_position[1])
        body.angle = angle
        body.dimensions = (width, length)
        body.angular_velocity = angular_velocity

        shape.elasticity = self.elasticity
        shape.friction = self.friction

        # add body and shape to space
        self.space.add(body, shape)

        # add body to agents dictionary
        self.bodies[body_id] = (body, shape)

    def update_body(self, body_id, specs):
        boundary = specs['boundary']
        length = boundary['length']
        width = boundary['width']
        mass = boundary['mass']
        thrust = boundary['thrust']
        torque = boundary['torque']

        body, shape = self.bodies[body_id]
        position = body.position
        angle = body.angle

        # get shape, inertia, make body, assign body to shape
        new_shape = self.get_shape(boundary)
        inertia = self.get_inertia(new_shape, mass)
        new_body = pymunk.Body(mass, inertia)
        new_shape.body = new_body

        new_body.position = position
        new_body.angle = angle
        new_body.velocity = body.velocity
        new_body.angular_velocity = body.angular_velocity
        new_body.dimensions = (width, length)
        new_body.thrust = thrust
        new_body.torque = torque

        new_shape.elasticity = shape.elasticity
        new_shape.friction = shape.friction

        # swap bodies
        self.space.remove(body, shape)
        self.space.add(new_body, new_shape)

        # update body
        self.bodies[body_id] = (new_body, new_shape)

    def update_bodies(self, bodies):
        # if an agent has been removed from the agents store,
        # remove it from space and bodies
        removed_bodies = [
            body_id for body_id in self.bodies.keys()
            if body_id not in bodies.keys()]
        for body_id in removed_bodies:
            body, shape = self.bodies[body_id]
            self.space.remove(body, shape)
            del self.bodies[body_id]

        # update agents, add new agents
        for body_id, specs in bodies.items():
            if body_id in self.bodies:
                self.update_body(body_id, specs)
            else:
                self.add_body_from_center(body_id, specs)

    def get_body_position(self, agent_id):
        body, shape = self.bodies[agent_id]
        return {
            'location': [pos for pos in body.position],
            'angle': body.angle,
        }

    def get_body_positions(self):
        return {
            body_id: {
                'boundary': self.get_body_position(body_id)}
            for body_id in self.bodies.keys()}



def test_multibody(
        total_time=2,
        agent_shape='rectangle',
        n_agents=1,
        jitter_force=1e1,
        screen=None):

    bounds = [500, 500]
    center_location = [0.5*loc for loc in bounds]
    agents = {
        str(agent_idx): {
            'boundary': {
                'location': center_location,
                'angle': random.uniform(0,2*PI),
                'volume': 15,
                'length': 30,
                'width': 10,
                'mass': 1,
                'thrust': 1e3,
                'torque': 0.0}}
        for agent_idx in range(n_agents)
    }
    config = {
        'agent_shape': agent_shape,
        'jitter_force': jitter_force,
        'bounds': bounds,
        'barriers': False,
        'initial_agents': agents,
        'screen': screen
    }
    multibody = PymunkMultibody(config)

    # run simulation
    time = 0
    time_step = 1
    while time < total_time:
        time += time_step
        multibody.run(time_step)


if __name__ == '__main__':
    test_multibody(10)



import numpy as np

from vivarium.library.units import Quantity, units
from vivarium.core.registry import updater_registry
from vivarium_cell.library.lattice_utils import (
    get_bin_site,
    get_bin_volume,
    count_to_concentration,
)

def update_field_with_exchange(current_value, new_value, states):
    '''Update environment with agent exchange

    Arguments:
        states: Dictionary with the following keys that specify the
            pre-update simulation state:

            * **global** (:py:class:`dict`): Contains the agent location
              under the ``location`` key.
            * **dimensions** (:py:class:`dict`): The contents of the
              environment's dimensions :term:`store` with the
              ``bounds``, ``n_bins``, and ``depth`` keys.

        new_value: Count of molecules to exchange with the environment
        current_value (numpy ndarray): The pre-update value of
            the field to update.

    Returns:
        The updated field.
    '''
    location = states['global']['location']
    n_bins = states['dimensions']['n_bins']
    bounds = states['dimensions']['bounds']
    depth = states['dimensions']['depth']
    exchange = new_value * units.count
    delta_field = np.zeros(
        (n_bins[0], n_bins[1]), dtype=np.float64)
    bin_site = get_bin_site(
        location, n_bins, bounds)
    bin_volume = get_bin_volume(
        n_bins, bounds, depth) * units.L
    concentration = count_to_concentration(exchange, bin_volume)
    delta_field[bin_site[0], bin_site[1]] += concentration.to(
        units.mmol / units.L).magnitude
    return current_value + delta_field

updater_registry.register(
    'update_field_with_exchange', update_field_with_exchange)



'''
Functions for making media
'''

from __future__ import absolute_import, division, print_function

import uuid

from parsimonious.grammar import Grammar
from parsimonious.nodes import NodeVisitor
from vivarium.library.units import units

# Raw data class
from vivarium_cell.data.knowledge_base import KnowledgeBase

INF = float("inf")
NEG_INF = float("-inf")

COUNTS_UNITS = units.mmol
VOLUME_UNITS = units.L
MASS_UNITS = units.g
CONC_UNITS = COUNTS_UNITS / VOLUME_UNITS

def cmp(a, b):
	return (a > b) - (a < b)

grammar = Grammar(r"""
    timeline = one_event*
    one_event = numeric? ws? recipe break?
    recipe = one_ingredient*
    one_ingredient = add? subtract? ingredient
    ingredient = media_id counts? volume? inf?
    counts = ws numeric count_key
    volume = ws numeric vol_key
    count_key = ws "mmol" ws?
    vol_key = ws "L" ws?
    inf = ws "Infinity" ws?
    media_id = ~"[A-Za-z0-9_][^,\s]+"i
    numeric = ~"[0-9.]+"i
    add = ws? "+" ws?
    subtract = ws? "-" ws?
    break = ws? "," ws?
    ws  = ~"\s*"
    """)

class AddIngredientsError(Exception):
    pass

class Media(object):
    '''
    A media object is a factory for making new media by combining
    media and ingredients at different volumes (with self.make_recipe().
    Ingredients can be added by either specifying their weight (in grams)
    or the counts (in mmol) in addition to the volume. The new media dicts
    are returned to the caller, and are not saved in this object. A media
    object holds dicts about stock media in ```self.stock_media``` and the
    formula weight of environmental molecules in
    ```self.environment_molecules_fw```, which is needed for mixing in
    ingredients at weights.
    '''

    def __init__(self):

        raw_data = KnowledgeBase()

        # get dicts from knowledge base
        self.environment_molecules_fw = self._get_environment_molecules_fw(raw_data)  # TODO (Eran) -- this assumes wcEcoli environment
        self.recipe_constructor = RecipeConstructor()
        self.recipes = self._get_recipes(raw_data)
        self._get_stock_media(raw_data)


    def _get_environment_molecules_fw(self, raw_data):
        '''get formula weight (units.g / units.mol) for all environmental molecules'''

        environment_molecules_fw = {}
        for row in raw_data.wcEcoli_environment_molecules:
            mol = row["molecule id"]
            fw = row["formula weight"]
            if fw == 'None':
                environment_molecules_fw[mol] = None
            else:
                environment_molecules_fw[mol] = float(fw) * (units.g / units.mol)

        return environment_molecules_fw

    def _get_stock_media(self, raw_data):
        '''load all stock media'''
        self.stock_media = {}

        for label in vars(raw_data.media):
            self.stock_media[label] = {}

            # get non-zero concentrations (assuming units.mmol / units.L)
            molecule_concentrations = getattr(raw_data.media, label)

            environment_non_zero_dict = {
                row["molecule id"]: row["concentration"]
                for row in molecule_concentrations}

            # update saved_media with non zero concentrations
            self.stock_media[label].update(environment_non_zero_dict)

        # add recipes to stock_media
        for media_id, recipe_raw in self.recipes.items():
            recipe_parsed = grammar.parse(recipe_raw)
            recipe = self.recipe_constructor.visit(recipe_parsed)
            media = self.make_recipe(recipe[0], True) # list only contains one recipe. Use first element.
            self.stock_media[media_id] = media

    def _get_recipes(self, raw_data):
        recipes = {}
        for row in raw_data.media_recipes:
            new_media_id = row["media id"]
            recipe = row["recipe"]
            recipes[new_media_id] = recipe
        return recipes

    def remove_units(self, media):
        return {mol: conc.m_as(CONC_UNITS) for mol, conc in media.items()}

    def get_saved_media(self, media_id, units=False):
        media = self.stock_media.get(media_id)
        if not units:
            media = self.remove_units(media)
        return media

    def add_media(self, media, new_media_id = str(uuid.uuid1())):
        '''add a media to stock'''

        # check/add units, if none defined, assume CONC_UNITS
        if all(isinstance(x, (float, int)) for x in media.values()):
            for mol_id, conc in media.items():
                media[mol_id] = conc * CONC_UNITS
        elif any(isinstance(x, (float, int)) for x in media.values()):
            raise AddIngredientsError(
                "inconsistent units for {}".format(new_media_id))

        self.stock_media[new_media_id] = media
        return new_media_id

    def make_recipe(self, recipe, units=False):
        '''make a single media recipe'''

        if recipe == 'end':
            return 'end'

        new_media = {}
        total_volume = 0.0 * VOLUME_UNITS
        for ingredient, amount in recipe.items():

            added_volume = amount.get('volume', 0.0 * VOLUME_UNITS)
            added_counts = amount.get('counts')
            added_weight = amount.get('weight')
            operation = amount.get('operation', 'add')

            added_media = {}
            if ingredient in self.stock_media:
                added_media = self.stock_media[ingredient]

            # if it is not an existing media, it needs weight or counts to add
            elif added_weight is not None:
                # added_weight takes priority over added_counts
                if added_weight.magnitude == INF:
                    added_conc = INF * CONC_UNITS
                elif added_weight.magnitude == 0.0:
                    added_conc = 0.0 * CONC_UNITS
                elif added_weight.magnitude < 0:
                    raise AddIngredientsError(
                        "Negative weight given for {}".format(ingredient))
                elif self.environment_molecules_fw[ingredient] is not None:
                    fw = self.environment_molecules_fw[ingredient]
                    added_counts = added_weight / fw
                    added_conc = added_counts / added_volume
                else:
                    raise AddIngredientsError(
                        "No fw defined for {} in wcEcoli_environment_molecules.tsv".format(ingredient))

                # save concentration
                added_media[ingredient] = added_conc

            elif added_counts is not None:
                # get new concentration
                # make infinite concentration of ingredient if mix_counts is Infinity
                if added_counts.magnitude == INF:
                    added_conc = INF * CONC_UNITS
                elif added_counts.magnitude == 0.0:
                    added_conc = 0.0 * CONC_UNITS
                elif added_counts.magnitude < 0:
                    raise AddIngredientsError(
                        "Negative counts given for {}".format(ingredient))
                else:
                    added_conc = added_counts / added_volume

                # save concentration
                added_media[ingredient] = added_conc

            else:
                raise AddIngredientsError(
                    "No added added weight or counts for {}".format(ingredient))

            if total_volume.magnitude == 0.0 and added_volume.magnitude == 0.0:
                # no volume, just merge media dicts. This is likely due to a call to a stock_media.
                new_media.update(added_media)
            else:
                new_media = self.combine_media(new_media, total_volume, added_media, added_volume, True, operation)

            total_volume += added_volume

        if not units:
            new_media = self.remove_units(new_media)

        return new_media

    def combine_media(self, media_1, volume_1, media_2, volume_2, units=False, operation='add'):

        # intialize new_media
        new_media = {mol_id: 0.0 * CONC_UNITS for mol_id in set(list(media_1.keys()) + list(media_2.keys()))}

        # get new_media volume
        new_volume = volume_1 + volume_2

        for mol_id in list(new_media.keys()):
            conc_1 = media_1.get(mol_id, 0 * CONC_UNITS)
            conc_2 = media_2.get(mol_id, 0 * CONC_UNITS)

            if conc_1.magnitude == INF or conc_2.magnitude == INF:
                new_conc = INF * CONC_UNITS
                if operation == 'subtract':
                    new_conc = 0.0 * CONC_UNITS
            else:
                counts_1 = conc_1 * volume_1
                counts_2 = conc_2 * volume_2
                new_counts = counts_1 + counts_2

                if operation == 'subtract':
                    new_counts = counts_1 - counts_2
                if new_counts.magnitude < 0:
                    raise AddIngredientsError(
                        "subtracting {} goes negative".format(mol_id))

                new_conc = new_counts / new_volume

            # update media
            new_media[mol_id] = new_conc

        if not units:
            new_media = self.remove_units(new_media)

        return new_media

    def make_timeline(self, timeline_str):
        '''
        Make a timeline from a string
        Args:
            timeline_str (str): 'time1 recipe1, time2 recipe2'
        Returns:
            timeline (list[tuple]): a list of tuples with (time (float), recipe (dict))
        '''

        timeline_parsed = grammar.parse(timeline_str)
        timeline_recipes = self.recipe_constructor.visit(timeline_parsed)
        timeline = []

        for time, recipe in timeline_recipes:
            media = self.make_recipe(recipe, True)
            if media == 'end':
                new_media_id = 'end'
            else:
                new_media_id = str(uuid.uuid1())
                # determine if this is an existing media
                for media_id, concentrations in self.stock_media.items():
                    if concentrations == media:
                    # if cmp(concentrations, media) == 0:
                        new_media_id = media_id
                        break
                self.stock_media[new_media_id] = media

            timeline.append((float(time), new_media_id))

        return timeline


class RecipeConstructor(NodeVisitor):
    '''
    Make a recipe from a parsed recipe expression.
    Args:
        - node: The node we're visiting
        - visited_children: The results of visiting the children of that node, in a list
    '''
    def visit_timeline(self, node, visited_children):
        timeline = []
        for child in visited_children:
            if child:
                timeline.append(child)
        return timeline

    def visit_one_event(self, node, visited_children):
        time, _, recipes, _ = visited_children

        if any(key in recipes for key in ['end', 'End', 'END']):
            event = (time[0], 'end')
        elif isinstance(time, list):
            event = (time[0], recipes)
        else:
            event = (recipes)
        return event

    def visit_recipe(self, node, visited_children):
        # put all added ingredients into a recipe dictionary
        recipe = {}
        for child in visited_children:
            recipe.update(child)
        return recipe

    def visit_one_ingredient(self, node, visited_children):
        add, subtract, ingredients = visited_children
        if isinstance(add, list):
            for amount in list(ingredients.values()):
                amount['operation'] = 'add'
        elif isinstance(subtract, list):
            for amount in list(ingredients.values()):
                amount['operation'] = 'subtract'
        return ingredients

    def visit_ingredient(self, node, visited_children):
        # put ingredients into small dicts, with their counts and volumes
        ingredient, counts, volume, infinity = visited_children
        if isinstance(counts, list) and isinstance(volume, list):
            recipe = {ingredient: {'counts': counts[0] * units.mmol, 'volume': volume[0] * units.L}}
        elif isinstance(volume, list):
            recipe = {ingredient: {'volume': volume[0] * units.L}}
        elif isinstance(infinity, list):
            recipe = {ingredient: {'counts': INF * units.L}}
        else:
            recipe = {ingredient: {}}
        return recipe

    def visit_counts(self, node, visited_children):
        return visited_children[1]

    def visit_volume(self, node, visited_children):
        return visited_children[1]

    def visit_media_id(self, node, visited_children):
        return (node.text)

    def visit_inf(self, node, visited_children):
        return INF

    def visit_numeric(self, node, visited_children):
        return float(node.text)

    def visit_add(self, node, visited_children):
        pass

    def visit_break(self, node, visited_children):
        pass

    def visit_ws(self, node, visited_children):
        pass

    def generic_visit(self, node, visited_children):
        # The generic visit method.
        return visited_children or node


def test_make_media():
    media_obj = Media()

    # Retrieve stock media
    media1 = media_obj.get_saved_media('M9_GLC', True)

    # Make media from ingredients
    ingredients = {
        'L-ALPHA-ALANINE': {'weight': 1.78 * units.g, 'volume': 0.025 * units.L},
        'ARG': {'weight': 8.44 * units.g, 'volume': 0.1 * units.L},
        'UREA': {'counts': 102.0 * units.mmol, 'volume': 1.0 * units.L},
        'LEU': {'weight': float("inf") * units.g, 'volume': 0 * units.L},
        'OXYGEN-MOLECULE': {'counts': 0 * units.g, 'volume': 0 * units.L},
    }
    media2 = media_obj.make_recipe(ingredients, True)

    # Combine two medias
    media3 = media_obj.combine_media(media1, 0.8 * units.L, media2, 0.2 * units.L)

    #Recipes with parsing expression grammer
    rc = RecipeConstructor()
    recipe_str = 'GLT 0.2 mmol 1 L + LEU 0.05 mmol .1 L + ARG 0.1 mmol .5 L'
    recipe_parsed = grammar.parse(recipe_str)
    recipe = rc.visit(recipe_parsed)
    media4 = media_obj.make_recipe(recipe[0])

    # Make timelines
    timeline_str = '0 minimal 1 L + GLT 0.2 mmol 1 L + LEU 0.05 mmol .1 L, ' \
                   '10 minimal_minus_oxygen 1 L + GLT 0.2 mmol 1 L, ' \
                   '100 minimal 1 L + GLT 0.2 mmol 1 L'
    timeline = media_obj.make_timeline(timeline_str)
    media1_id = timeline[0][1]
    media5 = media_obj.get_saved_media(media1_id)

if __name__ == '__main__':
    output = test_make_media()



from __future__ import absolute_import, division, print_function

import os
import csv
import random
from vivarium_cell.data.spreadsheets import load_tsv

FLAT_DIR = os.path.join('vivarium', 'data', 'flat')
MEDIA_IDS = ['minimal', 'minimal_minus_oxygen', 'minimal_plus_amino_acids']
FLUX_STR = 'lookup_flux_'

class LookUp(object):

    def __init__(self):

        self.lookup_avg = {media_id: {} for media_id in MEDIA_IDS}
        self.lookup_dist = {media_id: {} for media_id in MEDIA_IDS}

        for media_id in MEDIA_IDS:
            filename=os.path.join(FLAT_DIR, FLUX_STR + media_id + '.tsv')
            avg, dist = load_lookup(filename)
            self.lookup_avg[media_id].update(avg)
            self.lookup_dist[media_id].update(dist)

    def look_up(self, lookup_type, media, keys):
        ''' return look up values for each key in keys'''

        values = {}
        if lookup_type == 'average':
            values = {
                key: self.lookup_avg[media].get(key, None)
                for key in keys}
        if lookup_type == 'distribution':
            values = {
                key: random.choice(self.lookup_avg[media].get(key, None))
                for key in keys}

        return values


def load_lookup(filename):
    ''' load a file and pass back dicts with lookup_avg and lookup_dist'''
    lookup_avg = {}
    lookup_dist = {}

    rows = load_tsv(filename)
    for row in rows:
        key = row.get('id')
        avg = row.get('average')
        dist = row.get('distribution')

        # check if empty distribution
        if dist:
            dist = [float(value) for value in dist]
        else:
            dist=[0]

        lookup_avg[key] = avg
        lookup_dist[key] = dist

    return lookup_avg, lookup_dist



from __future__ import absolute_import, division, print_function

import os

import numpy as np

from vivarium.library.units import units
from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    plot_simulation_output,
    PROCESS_OUT_DIR,
)


NAME = 'growth'


class Growth(Process):
    """The Growth :term:`process class` models exponential cell growth.

    The cell's mass :math:`m_{t + h}` at time :math:`t + h` for
    :term:`timestep` :math:`h` and with growth rate :math:`r` is modeled
    as:

    .. math::

        m_{t + h} = m_t e^{rh}

    Configuration Options:

    * ``growth_rate``: The cell's growth rate :math:`r`. This rate is
      0.0006 by default.

      .. todo:: Why is the rate 0.0006?

    Example Usage:

    >>> import math
    >>> TIMESTEP = 1.0  # in seconds
    >>> # growth rate chosen so mass doubles each timestep
    >>> configuration = {'growth_rate': math.log(2.0)}
    >>> growth_process = Growth(configuration)
    >>> state = growth_process.default_state()
    >>> # the initial mass magnitude (femtograms)
    >>> mass = state['global']['mass'].magnitude
    >>> mass
    1339
    >>> update = growth_process.next_update(TIMESTEP, state)
    >>> # the new mass magnitude (femtograms)
    >>> new_mass = update['global']['mass'].magnitude
    >>> new_mass
    2678.0
    >>> new_mass / mass
    2.0

    """

    name = NAME
    defaults = {
        'growth_rate': 0.0006,
        'global_deriver_key': 'global_deriver',
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        self.growth_rate = self.or_default(initial_parameters, 'growth_rate')
        self.global_deriver_key = self.or_default(initial_parameters, 'global_deriver_key')
        parameters = {
            'growth_rate': self.growth_rate}
        parameters.update(initial_parameters)

        super(Growth, self).__init__(parameters)

    def ports_schema(self):
        return {
            'global': {
                'mass': {
                    '_emit': True,
                    '_default': 1339 * units.fg,
                    '_updater': 'set',
                    '_divider': 'split'},
                'volume': {
                    '_updater': 'set',
                    '_divider': 'split'},
                'divide': {
                    '_default': False,
                    '_updater': 'set'}}}

    def derivers(self):
        return {
            self.global_deriver_key: {
                'deriver': 'globals_deriver',
                'port_mapping': {
                    'global': 'global'},
                'config': {
                    'width': 1.0}}}

    def next_update(self, timestep, states):
        mass = states['global']['mass']
        new_mass = mass * np.exp(self.parameters['growth_rate'] * timestep)
        return {
            'global': {
                'mass': new_mass}}


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    growth = Growth({})
    settings = {'total_time': 10}
    timeseries = simulate_process_in_experiment(growth, settings)
    plot_simulation_output(timeseries, {}, out_dir)



'''
======================================
Expression Dotplot Colored by Survival
======================================
'''
from matplotlib import pyplot as plt
import numpy as np

from vivarium.core.experiment import get_in


PATH_TO_AGENTS = ('agents',)
PATH_TO_DEAD = ('boundary', 'dead')
LIVE_COLOR = 'red'
DEAD_COLOR = 'black'
ALPHA = 0.5


def plot_expression_survival(
    data, path_to_variable, xlabel, time_range=(0, 1)
):
    '''Create Expression Dotplot Colored by Survival

    Plot one dot for each cell along an axis to indicate that cell's
    average expression level for a specified protein. The dot color
    reflects whether the cell survived long enough to divide.

    Note that only the expression levels while the cell is alive are
    considered in the average.

    Parameters:
        data (dict): The raw data emitted from the simulation.
        path_to_variable (tuple): Path from the agent root to the
            variable that holds the protein's expression level. We do
            not adjust for cell volume, so this should be a
            concentration.
        xlabel (str): Label for x-axis.
        time_range (tuple): Tuple of two :py:class:`float`s that are
            fractions of the total simulated time period. These
            fractions indicate the start and end points (inclusive) of
            the time range to consider when calculating average
            expression level.

    Returns:
        plt.Figure: The finished figure.
    '''
    expression_levels = dict()
    die = set()
    end_time = max(data.keys())
    for time, time_data in data.items():
        if (time < time_range[0] * end_time
                or time > time_range[1] * end_time):
            continue
        agents_data = get_in(time_data, PATH_TO_AGENTS)
        for agent, agent_data in agents_data.items():
            lst = expression_levels.setdefault(agent, [])
            value = get_in(agent_data, path_to_variable)
            if value is not None:
                lst.append(value)
            if get_in(agent_data, PATH_TO_DEAD, False):
                die.add(agent)
            # Only count values when cell is alive
            elif value is not None:
                lst.append(value)

    live_averages = []
    dead_averages = []
    for agent, levels in expression_levels.items():
        if not levels:
            continue
        if agent in die:
            dead_averages.append(np.mean(levels))
        else:
            live_averages.append(np.mean(levels))

    fig, ax = plt.subplots(figsize=(6, 2))
    ax.scatter(
        live_averages, [0.1] * len(live_averages),
        label='Survive', color=LIVE_COLOR, alpha=ALPHA,
    )
    ax.scatter(
        dead_averages, [0.1] * len(dead_averages),
        label='Die', color=DEAD_COLOR, alpha=ALPHA,
    )
    ax.legend()
    ax.set_xlabel(xlabel)
    ax.set_ylim([0, 1.25])
    ax.get_yaxis().set_visible(False)
    for spine_name in ('left', 'top', 'right'):
        ax.spines[spine_name].set_visible(False)
    ax.spines['bottom'].set_position('zero')
    fig.tight_layout()
    return fig



'''
=====================
Colony Plotting Tools
=====================

This module contains tools to help you plot colony data.
'''

from __future__ import division, absolute_import, print_function

from matplotlib import pyplot as plt
import numpy as np
from scipy import stats

from vivarium.library.units import remove_units
from vivarium_cell.processes.derive_colony_shape import Variables


INCH_PER_COL = 4
INCH_PER_ROW = 2
SUBPLOT_W_SPACE = 0.4
SUBPLOT_H_SPACE = 1.5


#: Key for circumference in path timeseries
CIRCUMFERENCE_PATH = (Variables.CIRCUMFERENCE,)
#: Key for surface area in path timeseries
AREA_PATH = (Variables.AREA,)
#: Key for major axis in path timeseries
MAJOR_AXIS_PATH = (Variables.MAJOR_AXIS,)
#: Key for minor axis in path timeseries
MINOR_AXIS_PATH = (Variables.MINOR_AXIS,)
#: Key to which the cirumference-to-area ratio will be written in the
#: path timeseries
CIRCUMFERENCE_AREA_RATIO_PATH = 'circumference / surface_area'
#: Key to which the number of colonies is written in the path timeseries
NUM_COLONIES_PATH = 'Number of Colonies'
#: Key to which the ratio of major axis to minor axis is written in the
#: path timeseries
AXIS_RATIO_PATH = '(Major Axis) / (Minor Axis)'


# METRIC DERIVERS

def _derive_circumference_area_ratio(path_ts):
    if not (CIRCUMFERENCE_PATH in path_ts and AREA_PATH in path_ts):
        return
    circumference = path_ts[CIRCUMFERENCE_PATH]
    area = path_ts[AREA_PATH]
    ratio = [
        [
            c / a
            for c, a in zip(circumference_list, area_list)
        ]
        for circumference_list, area_list in zip(circumference, area)
    ]
    path_ts[CIRCUMFERENCE_AREA_RATIO_PATH] = ratio

def _derive_axis_ratio(path_ts):
    if not (MAJOR_AXIS_PATH in path_ts and MINOR_AXIS_PATH in path_ts):
        return
    major = path_ts[MAJOR_AXIS_PATH]
    minor = path_ts[MINOR_AXIS_PATH]
    ratio = [
        [
            major_val / minor_val
            for major_val, minor_val in zip(major_list, minor_list)
        ]
        for major_list, minor_list in zip(major, minor)
    ]
    path_ts[AXIS_RATIO_PATH] = ratio


#: List of metric derivers that will be applied to the path timeseries
_METRIC_DERIVERS = [
    _derive_circumference_area_ratio,
    _derive_axis_ratio,
]


def plot_colony_metrics(
    path_ts, title_size=16, tick_label_size=12, max_cols=5
):
    '''Plot colony metrics over time.

    Metric mean is plotted with SEM error bands.

    Arguments:
        path_ts (dict): Path timeseries of the data to plot. Each item
            in the dictionary should have as its key the path and as its
            value a list of values for each timepoint. Each value should
            be a list of metric values, one entry per colony. The
            dictionary should have one additional key, ``time``, whose
            value is a list of times for each timepoint.
        title_size (int): Font size for the title of each plot
        tick_label_size (int): Font size for each plot's axis tick
            labels.
        max_cols (int): The maximum number of columns. We add columns
            until we hit this limit, and only then do we add rows.

    Returns:
        matplotlib.figure.Figure: The plot as a Figure object.
    '''
    path_ts = remove_units(path_ts)
    for deriver in _METRIC_DERIVERS:
        deriver(path_ts)
    times = path_ts['time']
    del path_ts['time']
    # path_ts has tuples for keys. Here we turn those into strings so
    # that numpy doesn't iterate through the path elements
    path_ts = {
        str(key): val for key, val in path_ts.items()
    }
    arbitrary_metric = list(path_ts.keys())[0]
    path_ts[NUM_COLONIES_PATH] = [
        len(timepoint) for timepoint in path_ts[arbitrary_metric]
    ]
    # Create Figure
    paths = sorted(path_ts.keys())
    n_cols = min(len(paths), max_cols)
    n_rows = int(np.ceil(len(paths) / n_cols))
    fig = plt.figure(
        figsize=(INCH_PER_COL * n_cols, INCH_PER_ROW * n_rows))
    grid = plt.GridSpec(
        ncols=n_cols, nrows=n_rows, wspace=SUBPLOT_W_SPACE,
        hspace=SUBPLOT_H_SPACE
    )

    # Assign paths to subplot coordinates
    padding = [None] * int(n_cols * n_rows - len(paths))
    paths += padding
    paths_grid = np.array(paths)
    paths_grid = paths_grid.reshape((n_rows, n_cols))

    # Create the subplots
    for i in range(n_rows):
        for j in range(n_cols):
            path = paths_grid[i, j]
            if path is None:
                continue
            ax = fig.add_subplot(grid[i, j])
            # Configure axes and titles
            for tick_type in ('major', 'minor'):
                ax.tick_params(
                    axis='both', which=tick_type,
                    labelsize=tick_label_size
                )
            ax.title.set_text(path)
            ax.title.set_fontsize(title_size)
            ax.set_xlim([times[0], times[-1]])
            ax.xaxis.get_offset_text().set_fontsize(tick_label_size)
            ax.yaxis.get_offset_text().set_fontsize(tick_label_size)
            ax.set_xlabel('time (s)', fontsize=title_size)
            # Plot data
            data = path_ts[path]
            if path == NUM_COLONIES_PATH:
                ax.plot(times, data)
            else:
                means = []
                sems = []
                plot_times = []
                for i_time, metrics_list in enumerate(data):
                    if not metrics_list:
                        continue
                    array = np.array(metrics_list)
                    means.append(np.mean(array))
                    sems.append(
                        stats.sem(array) if len(array) > 1 else 0)
                    plot_times.append(times[i_time])
                x = np.array(plot_times)
                y = np.array(means)
                yerr = np.array(sems)
                yerr[np.isnan(yerr)] = 0
                ax.plot(x, y)
                ax.fill_between(x, y - yerr, y + yerr, alpha=0.2)

    return fig


def plot_metric_across_experiments(
    path_ts_dict, path, title=None, xlabel='time (s)', ylabel=None,
    title_size=16, tick_label_size=12,
):
    '''Overlay plots of a single metric from different experiments.

    Parameters:
        path_ts_dict (dict): Map from the string to use as the label for
            the experiment in the legend to that experiment's path
            timeseries.
        path (tuple): Path to plot. Should be a key in each value of
            ``path_ts_dict``.
        title (str): Plot title. If None, no title is set.
        xlabel (str): X-axis label. If None, no label is set.
        ylabel (str): Y-axis label. If None, no label is set.
        title_size (float): Font size for plot and axis titles.
        tick_label_size (float): Font size for tick labels.

    Returns:
        The figure with the plot.
    '''
    fig, ax = plt.subplots()

    # Set labels and font sizes
    if title is not None:
        ax.set_title(title, fontsize=title_size)
    if xlabel is not None:
        ax.set_xlabel(xlabel, fontsize=title_size)
    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=title_size)
    ax.xaxis.get_offset_text().set_fontsize(tick_label_size)
    ax.yaxis.get_offset_text().set_fontsize(tick_label_size)

    # Plot data
    for label, path_ts in path_ts_dict.items():
        data = path_ts[path]
        times = path_ts['time']
        if path == NUM_COLONIES_PATH:
            ax.plot(times, data, label=label)
        else:
            plot_times = []
            means = []
            for i, metrics_list in enumerate(data):
                if not metrics_list:
                    continue
                means.append(np.mean(metrics_list))
                plot_times.append(times[i])
            ax.plot(plot_times, means, label=label)

    ax.legend()
    fig.tight_layout()
    return fig



from __future__ import absolute_import, division, print_function

import numpy as np
from arrow import StochasticSystem


from vivarium.core.process import Process
from vivarium.core.composition import simulate_process_in_experiment

from vivarium.library.dict_utils import keys_list
from vivarium_cell.data.molecular_weight import molecular_weight
from vivarium_cell.data.chromosomes.flagella_chromosome import FlagellaChromosome

chromosome = FlagellaChromosome()

def build_complexation_stoichiometry(
        stoichiometry,
        rates,
        reaction_ids,
        monomer_ids,
        complex_ids):

    molecule_ids = monomer_ids + complex_ids
    matrix = np.zeros((len(stoichiometry), len(molecule_ids)), dtype=np.int64)
    rates_array = np.zeros(len(stoichiometry))

    reverse_index = {
        molecule_id: index
        for index, molecule_id in enumerate(molecule_ids)}

    for reaction_index, reaction_id in enumerate(reaction_ids):
        reaction = stoichiometry[reaction_id]
        rates_array[reaction_index] = rates[reaction_id]
        for molecule_id, level in reaction.items():
            matrix[reaction_index][reverse_index[molecule_id]] = level

    return matrix, rates_array


class Complexation(Process):

    name = 'complexation'
    defaults = {
        'monomer_ids': chromosome.complexation_monomer_ids,
        'complex_ids': chromosome.complexation_complex_ids,
        'stoichiometry': chromosome.complexation_stoichiometry,
        'rates': chromosome.complexation_rates,
        'mass_deriver_key': 'mass_deriver',
        'time_step': 1.0,
    }

    def __init__(self, initial_parameters=None):
        if not initial_parameters:
            initial_parameters = {}

        super(Complexation, self).__init__(initial_parameters)

        self.derive_defaults('stoichiometry', 'reaction_ids', keys_list)

        self.monomer_ids = self.parameters['monomer_ids']
        self.complex_ids = self.parameters['complex_ids']
        self.reaction_ids = self.parameters['reaction_ids']

        self.stoichiometry = self.parameters['stoichiometry']
        self.rates = self.parameters['rates']

        self.complexation_stoichiometry, self.complexation_rates = build_complexation_stoichiometry(
            self.stoichiometry,
            self.rates,
            self.reaction_ids,
            self.monomer_ids,
            self.complex_ids)

        self.complexation = StochasticSystem(self.complexation_stoichiometry)

        self.mass_deriver_key = self.or_default(initial_parameters, 'mass_deriver_key')


    def ports_schema(self):
        return {
            'monomers': {
                monomer: {
                    '_default': 0,
                    '_emit': True,
                    '_divider': 'split',
                    '_properties': {
                        'mw': molecular_weight[
                            monomer]} if monomer in molecular_weight else {}}
                for monomer in self.monomer_ids},
            'complexes': {
                complex: {
                    '_default': 0,
                    '_emit': True,
                    '_divider': 'split',
                    '_properties': {
                        'mw': molecular_weight[
                            complex]} if complex in molecular_weight else {}}
                for complex in self.complex_ids},
            'global': {}}

    def derivers(self):
        return {
            self.mass_deriver_key: {
                'deriver': 'mass_deriver',
                'port_mapping': {
                    'global': 'global'}}}

    def next_update(self, timestep, states):
        monomers = states['monomers']
        complexes = states['complexes']

        substrate = np.zeros(len(self.monomer_ids) + len(self.complex_ids), dtype=np.int64)

        for index, monomer_id in enumerate(self.monomer_ids):
            substrate[index] = monomers[monomer_id]
        for index, complex_id in enumerate(self.complex_ids):
            substrate[index + len(self.monomer_ids)] = complexes[complex_id]

        try:
            result = self.complexation.evolve(
                timestep,
                substrate,
                self.complexation_rates)
        except:
            print('Failed simulation. \n substrate {} \n monomers {} \n complexes {}'.format(
                substrate, self.monomer_ids, self.complex_ids))

        outcome = result['outcome'] - substrate

        monomers_update = {
            monomer_id: outcome[index]
            for index, monomer_id in enumerate(self.monomer_ids)}

        complexes_update = {
            complex_id: outcome[index + len(self.monomer_ids)]
            for index, complex_id in enumerate(self.complex_ids)}

        update = {
            'monomers': monomers_update,
            'complexes': complexes_update}

        return update

def test_complexation():
    complexation = Complexation()
    state = {
        'monomers': {
            monomer: 1000
            for monomer in complexation.monomer_ids},
        'complexes': {
            complex: 0
            for complex in complexation.complex_ids}}

    settings = {
        'total_time': 10,
        'initial_state': state,
    }
    data = simulate_process_in_experiment(complexation, settings)

    print(data)


if __name__ == '__main__':
    test_complexation()



from __future__ import absolute_import, division, print_function

from vivarium.library.units import units
from vivarium.core.process import Process



class DivisionVolume(Process):

    name = 'division_volume'
    defaults = {
        'initial_state': {},
        'division_volume': 2.4 * units.fL,  # fL
    }

    def __init__(self, initial_parameters=None):
        if not initial_parameters:
            initial_parameters = {}

        self.division = 0
        division_volume = initial_parameters.get('division_volume', self.defaults['division_volume'])

        parameters = {
            'division_volume': division_volume}  # TODO -- make division at 2X initial_volume?  Pass this in from initial_parameters

        super(DivisionVolume, self).__init__(parameters)

    def ports_schema(self):
        return {
            'global': {
                'divide': {
                    '_default': False,
                    '_emit': True,
                    '_updater': 'set',
                    '_divider': 'zero'},
                'volume': {
                    '_default': 1.2 * units.fL}}}

    def next_update(self, timestep, states):
        volume = states['global']['volume']
        if volume >= self.parameters['division_volume']:
            self.division = True
            return {'global': {'divide': self.division}}
        else:
            return {}



"""
===================================
Toy Glucose Phosphorylation Process
===================================

This is a toy example referenced in the documentation.
"""

from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process,
    plot_simulation_output,
)
from vivarium.library.units import units


class GlucosePhosphorylation(Process):

    name = 'glucose_phosphorylation'
    defaults = {
        'k_cat': 2e-3,
        'K_ATP': 5e-2,
        'K_GLC': 4e-2,
    }

    def __init__(self, initial_parameters=None):
        ports = {
            'nucleoside_phosphates': ['ATP', 'ADP'],
            'cytoplasm': ['GLC', 'G6P', 'HK'],
            'global': ['mass']
        }
        parameters = GlucosePhosphorylation.defaults
        parameters.update(initial_parameters)
        super(GlucosePhosphorylation, self).__init__(
            ports, parameters)

    def next_update(self, timestep, states):
        # Get concentrations from state
        cytoplasm = states['cytoplasm']
        nucleoside_phosphates = states['nucleoside_phosphates']
        hk = cytoplasm['HK']
        glc = cytoplasm['GLC']
        atp = nucleoside_phosphates['ATP']

        # Get kinetic parameters
        k_cat = self.parameters['k_cat']
        k_atp = self.parameters['K_ATP']
        k_glc = self.parameters['K_GLC']

        # Compute reaction rate with michaelis-menten equation
        rate = k_cat * hk * glc * atp / (
            k_glc * k_atp + k_glc * atp + k_atp * glc + glc * atp)

        # Compute concentration changes from rate and timestep
        delta_glc = -rate * timestep
        delta_atp = -rate * timestep
        delta_g6p = rate * timestep
        delta_adp = rate * timestep

        # Compile changes into an update
        update = {
            'cytoplasm': {
                'GLC': delta_glc,
                'G6P': delta_g6p,
                # We exclude HK because it doesn't change
            },
            'nucleoside_phosphates': {
                'ATP': delta_atp,
                'ADP': delta_adp,
            },
        }

        return update

    def ports_schema(self):
        return {
            'cytoplasm': {
                'GLC': {
                    # accumulate means to add the updates
                    '_updater': 'accumulate',
                    '_default': 1.0,
                    '_properties': {
                        'mw': 1.0 * units.g / units.mol,
                    },
                    '_emit': True,
                },
                # accumulate is the default, so we don't need to specify
                # updaters for the rest of the variables
                'G6P': {
                    '_default': 0.0,
                    '_properties': {
                        'mw': 1.0 * units.g / units.mol,
                    },
                    '_emit': True,
                },
                'HK': {
                    '_default': 0.1,
                    '_properties': {
                        'mw': 1.0 * units.g / units.mol,
                    },
                },
            },
            'nucleoside_phosphates': {
                'ATP': {
                    '_default': 2.0,
                    '_emit': True,
                },
                'ADP': {
                    '_default': 0.0,
                    '_emit': True,
                }
            },
            'global': {
            },
        }

    def derivers(self):
        return {
            'my_deriver': {
                'deriver': 'mass_deriver',
                'port_mapping': {
                    'global': 'global',
                },
                'config': {},
            },
        }


if __name__ == '__main__':
    parameters = {
        'k_cat': 1.5,
    }
    my_process = GlucosePhosphorylation(parameters)

    settings = {
        'total_time': 10,
        'timestep': 0.1,
    }
    timeseries = simulate_process(my_process, settings)
    plot_simulation_output(timeseries, {}, './')



from __future__ import absolute_import, division, print_function

import os

from vivarium.core.process import Generator
from vivarium.core.composition import (
    COMPARTMENT_OUT_DIR,
    simulate_compartment_in_experiment,
    plot_simulation_output,
)
from vivarium.library.units import units

# processes
from vivarium.processes.tree_mass import TreeMass

from vivarium_cell.plots.gene_expression import plot_gene_expression_output
from vivarium_cell.processes.transcription import Transcription, UNBOUND_RNAP_KEY
from vivarium_cell.processes.translation import Translation, UNBOUND_RIBOSOME_KEY
from vivarium_cell.processes.degradation import RnaDegradation
from vivarium_cell.processes.complexation import Complexation
from vivarium_cell.processes.division_volume import DivisionVolume
from vivarium_cell.data.amino_acids import amino_acids
from vivarium_cell.data.nucleotides import nucleotides
from vivarium_cell.states.chromosome import toy_chromosome_config


NAME = 'gene_expression'

class GeneExpression(Generator):

    defaults = {
        'global_path': ('global',),
        'initial_mass': 1339.0 * units.fg,
        'time_step': 1.0,
        'transcription': {},
        'translation': {},
        'degradation': {},
        'complexation': {},
    }

    def __init__(self, config):
        super(GeneExpression, self).__init__(config)

    def generate_processes(self, config):
        transcription_config = config['transcription']
        translation_config = config['translation']
        degradation_config = config['degradation']
        complexation_config = config['complexation']

        # update timestep
        transcription_config.update({'time_step': config['time_step']})
        translation_config.update({'time_step': config['time_step']})
        degradation_config.update({'time_step': config['time_step']})
        complexation_config.update({'time_step': config['time_step']})

        # make the processes
        transcription = Transcription(transcription_config)
        translation = Translation(translation_config)
        degradation = RnaDegradation(degradation_config)
        complexation = Complexation(complexation_config)
        mass_deriver = TreeMass(config.get('mass_deriver', {
            'initial_mass': config['initial_mass']}))
        division = DivisionVolume(config)

        return {
            'mass_deriver': mass_deriver,
            'transcription': transcription,
            'translation': translation,
            'degradation': degradation,
            'complexation': complexation,
            'division': division}

    def generate_topology(self, config):
        global_path = config['global_path']

        return {
            'mass_deriver': {
                'global': global_path},

            'transcription': {
                'chromosome': ('chromosome',),
                'molecules': ('molecules',),
                'proteins': ('proteins',),
                'transcripts': ('transcripts',),
                'factors': ('concentrations',),
                'global': global_path},

            'translation': {
                'ribosomes': ('ribosomes',),
                'molecules': ('molecules',),
                'transcripts': ('transcripts',),
                'proteins': ('proteins',),
                'concentrations': ('concentrations',),
                'global': global_path},

            'degradation': {
                'transcripts': ('transcripts',),
                'proteins': ('proteins',),
                'molecules': ('molecules',),
                'global': global_path},

            'complexation': {
                'monomers': ('proteins',),
                'complexes': ('proteins',),
                'global': global_path},

            'division': {
                'global': global_path}}


# test
def run_gene_expression(total_time=10, out_dir='out'):
    timeseries = test_gene_expression(total_time)
    plot_settings = {
        'name': 'gene_expression',
        'ports': {
            'transcripts': 'transcripts',
            'molecules': 'molecules',
            'proteins': 'proteins'}}
    plot_gene_expression_output(timeseries, plot_settings, out_dir)

    sim_plot_settings = {'max_rows': 25}
    plot_simulation_output(timeseries, sim_plot_settings, out_dir)

def test_gene_expression(total_time=10):
    # load the compartment
    compartment_config = {
        'external_path': ('external',),
        'global_path': ('global',),
        'agents_path': ('..', '..', 'cells',),
        'transcription': {
            'sequence': toy_chromosome_config['sequence'],
            'templates': toy_chromosome_config['promoters'],
            'genes': toy_chromosome_config['genes'],
            'promoter_affinities': toy_chromosome_config['promoter_affinities'],
            'transcription_factors': ['tfA', 'tfB'],
            'elongation_rate': 10.0},
        # 'complexation': {
        #     'monomer_ids': [],
        #     'complex_ids': [],
        #     'stoichiometry': {}}
    }
    compartment = GeneExpression(compartment_config)

    molecules = {
        nt: 1000
        for nt in nucleotides.values()}
    molecules.update({
        aa: 1000
        for aa in amino_acids.values()})

    proteins = {
        polymerase: 100
        for polymerase in [
                UNBOUND_RNAP_KEY,
                UNBOUND_RIBOSOME_KEY]}

    proteins.update({
        factor: 1
        for factor in [
                'tfA',
                'tfB']})

    # simulate
    settings = {
        'timestep': 1,
        'total_time': total_time,
        'initial_state': {
            'proteins': proteins,
            'molecules': molecules}}
    return simulate_compartment_in_experiment(compartment, settings)


if __name__ == '__main__':
    out_dir = os.path.join(COMPARTMENT_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    run_gene_expression(600, out_dir)



from __future__ import absolute_import, division, print_function

import os
import random

from vivarium.core.process import Process
from vivarium.library.dict_utils import tuplify_port_dicts
from vivarium.core.composition import (
    simulate_process_in_experiment,
    plot_simulation_output,
    PROCESS_OUT_DIR,
)

from vivarium_cell.library.regulation_logic import build_rule


NAME = 'minimal_expression'


class MinimalExpression(Process):
    '''
    a minimal protein expression process.
    TO BE USED ONLY AS TRAINING WHEELS

    parameters:
        expression_rates (dict) with {'mol_id': probability_of_expression (1/sec)}
    '''

    name = NAME
    defaults = {
        'step_size': 1,
        'regulation': {},
        'concentrations_deriver_key': 'concentrations_deriver',
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        expression_rates = initial_parameters.get('expression_rates')
        self.internal_states = list(expression_rates.keys()) if expression_rates else []
        regulation_logic = initial_parameters.get('regulation', self.defaults['regulation'])
        self.regulation = {
            gene_id: build_rule(logic) for gene_id, logic in regulation_logic.items()}
        regulators = initial_parameters.get('regulators', [])
        self.internal_regulators = [state_id for port_id, state_id in regulators if port_id == 'internal']
        self.external_regulators = [state_id for port_id, state_id in regulators if port_id == 'external']

        parameters = {
            'expression_rates': expression_rates,
            'step_size': initial_parameters.get('step_size', self.defaults['step_size'])}
        parameters.update(initial_parameters)

        self.concentrations_deriver_key = self.or_default(initial_parameters, 'concentrations_deriver_key')

        super(MinimalExpression, self).__init__(parameters)

    def ports_schema(self):
        return {
            'global': {},
            'concentrations': {},
            'external': {
                state: {
                    '_default': 0.0}
                for state in self.external_regulators},
            'internal': {
                state: {
                    '_default': 0.0,
                    '_emit': True}
                for state in self.internal_states + self.internal_regulators}}

    def derivers(self):
        return {
            self.concentrations_deriver_key: {
                'deriver': 'concentrations_deriver',
                'port_mapping': {
                    'global': 'global',
                    'counts': 'internal',
                    'concentrations': 'concentrations'},
                'config': {
                    'concentration_keys': self.internal_states + self.internal_regulators}}}

    def next_update(self, timestep, states):
        internal = states['internal']
        step_size = self.parameters['step_size']
        n_steps = int(timestep / step_size)

        # get state of regulated reactions (True/False)
        flattened_states = tuplify_port_dicts(states)
        regulation_state = {}
        for gene_id, reg_logic in self.regulation.items():
            regulation_state[gene_id] = reg_logic(flattened_states)

        internal_update = {state_id: 0 for state_id in internal.keys()}
        for state_id in internal.keys():
            if state_id in regulation_state and not regulation_state[state_id]:
                break
            rate = self.parameters['expression_rates'][state_id]
            for step in range(n_steps):
                if random.random() < rate:
                    internal_update[state_id] += 1

        return {
            'internal': internal_update}



# test functions
def get_toy_expression_config():
    toy_expression_rates = {
        'protein1': 1e-2,
        'protein2': 1e-1,
        'protein3': 1e0}

    return {
        'expression_rates': toy_expression_rates}

def test_expression(end_time=10):
    expression_config = get_toy_expression_config()
    # load process
    expression = MinimalExpression(expression_config)
    settings = {'total_time': end_time}
    return simulate_process_in_experiment(expression, settings)



if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    timeseries = test_expression(1000)
    plot_simulation_output(timeseries, {}, out_dir)




from __future__ import absolute_import, division, print_function

import copy
import os

from vivarium.core.process import Generator
from vivarium.core.composition import (
    compartment_in_experiment,
    COMPARTMENT_OUT_DIR,
)
from vivarium.library.dict_utils import deep_merge

# processes
from vivarium_cell.processes.multibody_physics import (
    Multibody,
    agent_body_config,
)
from vivarium_cell.processes.diffusion_field import (
    DiffusionField,
    get_gaussian_config,
)
from vivarium_cell.processes.derive_colony_shape import ColonyShapeDeriver

# plots
from vivarium_cell.plots.multibody_physics import plot_snapshots


NAME = 'lattice'


# make a configuration dictionary for the Lattice compartment
def make_lattice_config(
        time_step=None,
        jitter_force=None,
        bounds=None,
        n_bins=None,
        depth=None,
        concentrations=None,
        molecules=None,
        diffusion=None,
        keep_fields_emit=None,
        set_config=None,
        parallel=None,
):
    config = {'multibody': {}, 'diffusion': {}}

    if time_step:
        config['multibody']['time_step'] = time_step
        config['diffusion']['time_step'] = time_step
    if bounds:
        config['multibody']['bounds'] = bounds
        config['diffusion']['bounds'] = bounds
        config['diffusion']['n_bins'] = bounds
    if n_bins:
        config['diffusion']['n_bins'] = n_bins
    if jitter_force:
        config['multibody']['jitter_force'] = jitter_force
    if depth:
        config['diffusion']['depth'] = depth
    if diffusion:
        config['diffusion']['diffusion'] = diffusion
    if concentrations:
        config['diffusion']['gradient'] = {
            'type': 'uniform',
            'molecules': concentrations}
        molecules = list(concentrations.keys())
        config['diffusion']['molecules'] = molecules
    elif molecules:
        # molecules are a list, assume uniform concentrations of 1
        config['diffusion']['molecules'] = molecules
    if keep_fields_emit:
        # by default no fields are emitted
        config['diffusion']['_schema'] = {
            'fields': {
                field_id: {
                    '_emit': False}
                for field_id in molecules
                if field_id not in keep_fields_emit}}
    if parallel:
        config['diffusion']['_parallel'] = True
        config['multibody']['_parallel'] = True
    if set_config:
        config = deep_merge(config, set_config)

    return config



class Lattice(Generator):
    """
    Lattice:  A two-dimensional lattice environmental model with multibody physics and diffusing molecular fields.
    """

    defaults = {
        # To exclude a process, from the compartment, set its
        # configuration dictionary to None, e.g. colony_mass_deriver
        'multibody': {
            'bounds': [10, 10],
            'size': [10, 10],
        },
        'diffusion': {
            'molecules': ['glc'],
            'n_bins': [10, 10],
            'size': [10, 10],
            'depth': 3000.0,
            'diffusion': 1e-2,
        },
        'colony_shape_deriver': None,
        '_schema': {},
    }

    def __init__(self, config=None):
        super(Lattice, self).__init__(config)

    def generate_processes(self, config):
        processes = {
            'multibody': Multibody(config['multibody']),
            'diffusion': DiffusionField(config['diffusion'])
        }
        colony_shape_config = config['colony_shape_deriver']
        if colony_shape_config is not None:
            processes['colony_shape_deriver'] = ColonyShapeDeriver(
                colony_shape_config)
        return processes

    def generate_topology(self, config):
        topology = {
            'multibody': {
                'agents': ('agents',),
            },
            'diffusion': {
                'agents': ('agents',),
                'fields': ('fields',),
                'dimensions': ('dimensions',),
            },
            'colony_shape_deriver': {
                'colony_global': ('colony_global',),
                'agents': ('agents',),
            }
        }
        return {
            process: process_topology
            for process, process_topology in topology.items()
            if config[process] is not None
        }


def test_lattice(
        config=None,
        n_agents=1,
        end_time=10
):
    if config is None:
        config = make_lattice_config()
    # configure the compartment
    compartment = Lattice(config)

    # set initial agent state
    if n_agents:
        agent_ids = [str(agent_id) for agent_id in range(n_agents)]
        body_config = {'agent_ids': agent_ids}
        if 'multibody' in config and 'bounds' in config['multibody']:
            body_config.update({'bounds': config['multibody']['bounds']})
        initial_agents_state = agent_body_config(body_config)
        initial_state = {'agents': initial_agents_state}

    # configure experiment
    experiment_settings = {
        'compartment': config,
        'initial_state': initial_state}
    experiment = compartment_in_experiment(
        compartment,
        experiment_settings)

    # run experiment
    timestep = 1
    time = 0
    while time < end_time:
        experiment.update(timestep)
        time += timestep
    data = experiment.emitter.get_data()

    # assert that the agent remains in the simulation until the end
    assert len(data[end_time]['agents']) == n_agents
    return data


def main():
    out_dir = os.path.join(COMPARTMENT_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    bounds = [25, 25]
    config = make_lattice_config(
        bounds=bounds,
    )
    data = test_lattice(
        config=config,
        n_agents=1,
        end_time=40)

    # make snapshot plot
    agents = {time: time_data['agents'] for time, time_data in data.items()}
    fields = {time: time_data['fields'] for time, time_data in data.items()}
    plot_data = {
        'agents': agents,
        'fields': fields,
        'config': {'bounds': bounds}}
    plot_config = {
        'out_dir': out_dir,
        'filename': 'snapshots'}
    plot_snapshots(plot_data, plot_config)


if __name__ == '__main__':
    main()



'''
Rate law utilities

This collection of functions can assist in constructing and analyzing kinetic rate laws
that are generated in wholecell.kinetic_rate_laws.kinetic_rate_laws.

Functions include:
    - load_reactions(): returns a dict of all_reactions, with location tags added to the enzymes
    - get_reactions_from_exchange: provided a dict of reactions and exchange molecules, this returns a list of all reactions for those exchange molecules
    - get_molecules_from_stoich: given a list of desired reactions and dict of all reactions with stoichiometry, returns all the relevant molecules -- substrates and enzymes

The RateLawUtilities module can be called with:
> python -m vivarium.library.rate_law_utilities

'''

from __future__ import absolute_import, division, print_function

import os
import json
import numpy as np
import argparse

# import matplotlib.pyplot as plt  # RuntimeError: Python is not installed as a framework.
import matplotlib.pyplot as plt

from vivarium.library import filepath

from vivarium_cell.data.spreadsheets import load_tsv
from vivarium_cell.library.look_up import LookUp
import vivarium_cell.library.kinetic_rate_laws as rate_laws

REACTIONS_FILE = os.path.join("vivarium", "data", "flat", "wcEcoli_reactions.tsv")
PROTEINS_FILE = os.path.join("vivarium", "data", "flat", "wcEcoli_proteins.tsv")
COMPLEXATION_FILE = os.path.join("vivarium", "data", "flat", "wcEcoli_complexationReactions.tsv")
KINETIC_PARAMETERS_PATH = os.path.join("vivarium", 'data', 'json_files')
OUTPUT_DIR = os.path.join('out', 'kinetic_rate_laws')


def analyze_rate_laws(kinetic_rate_laws, baseline_concentrations, output_filename):
    '''
    Args:
        kinetic_rate_laws (object): a configured kinetic_rate_law object
        baseline_concentrations (dict): concentrations for all molecules required for the rate laws

    Function:
        Runs an analysis of all rate laws in kinetic_rate_laws and saves the output in a plot

    '''

    plot_text = True
    test_transporter = True
    test_cofactor = True
    test_competitor = True

    reactions = kinetic_rate_laws.reactions
    kinetic_parameters = kinetic_rate_laws.kinetic_parameters
    rate_law_configuration = kinetic_rate_laws.rate_law_configuration

    ## Plot analysis
    n_samples = 100
    n_samples_shown = 10
    n_rxns = len(reactions)

    # get concentrations array for testing and plotting
    conc_samples = np.logspace(-8, 1, num=n_samples, endpoint=True, base=10)
    conc_shown = np.flip(np.logspace(-8, 1, num=n_samples_shown, endpoint=True, base=10), 0) # use reverse ordering for legend order
    transporter_concs_shown = np.flip(np.logspace(-4, 1, num=n_samples_shown, endpoint=True, base=10), 0) # use reverse ordering for legend order

    # TODO -- should count number of transporters in each, because each gets an extra row
    rows = sum([plot_text, test_cofactor, test_transporter, test_competitor]) * n_rxns + n_rxns
    columns = 3


    cmap = plt.cm.get_cmap('Spectral')
    colors = [cmap(float(idx) / n_samples_shown) for idx in reversed(range(n_samples_shown))]  # use reverse ordering for legend order

    plt.figure(figsize=(6*columns, 3*rows))
    row_number = 0
    col_number = 0

    for reaction_id, specs in reactions.items():
        transporters = specs.get('catalyzed by')
        stoich = specs.get('stoichiometry')
        parameters = kinetic_parameters.get(reaction_id)

        reactants = [mol for mol, coeff in stoich.items() if coeff < 0]
        products = [mol for mol, coeff in stoich.items() if coeff > 0]

        if plot_text:
            plt.subplot(rows, columns, row_number * columns + col_number + 1)
            plt.text(0.02, 0.6, 'reaction: %s' % reaction_id, weight='bold', fontsize=14)
            plt.text(0.02, 0.45, 'reactants: %s' % reactants, fontsize=14)
            plt.text(0.02, 0.3, 'products: %s' % products, fontsize=14)
            plt.text(0.02, 0.15, 'transporters: %s' % transporters, fontsize=14)
            plt.text(0.02, 0.0, 'parameters: %s' % parameters, fontsize=14, wrap=True)
            plt.axis('off')

            row_number += 1

        # test rate law by sampling substrate concentrations
        for transporter in transporters:
            if transporter not in parameters:
                continue

            # get cofactors list
            analyze_cofactors = []
            for reactant in reactants:
                if parameters[transporter][reactant] is not None:
                    analyze_cofactors.append(reactant)

            # use first listed cofactor as basal TODO -- pass in cofactor of interest
            a1 = analyze_cofactors[0]

            # get competitors list
            rxns_transporter = rate_law_configuration[transporter]['reaction_cofactors'].keys()
            competing_rxns = [trpr for trpr in rxns_transporter if trpr not in reaction_id]

            analyze_competitors = []
            for rx in competing_rxns:
                competitor_candidates = rate_law_configuration[transporter]['reaction_cofactors'][rx]
                competing_mols = [mol for mol in competitor_candidates[0] if mol not in analyze_cofactors]
                analyze_competitors.extend(competing_mols)

            if test_cofactor:
                for cofactor in analyze_cofactors:

                    concentrations = baseline_concentrations.copy()

                    if cofactor is a1:
                        # plot a1 alone

                        flux_values = scan_conc(kinetic_rate_laws, concentrations, reaction_id, a1, conc_samples)

                        # plot M-M curve for this reaction
                        plt.subplot(rows, columns, row_number * columns + col_number + 1)
                        plt.plot(conc_samples, flux_values)

                        plt.xscale('log')
                        plt.xlabel('%s concentration (M)' % cofactor)
                        plt.ylabel('flux (M/s)')
                        plt.title('cofactor: %s' % cofactor)

                        col_number += 1

                    else:
                        # plot affect on a1 of varying cofactor
                        plt.subplot(rows, columns, row_number * columns + col_number + 1)
                        for index, cofactor_conc in enumerate(conc_shown):
                            concentrations[cofactor] = cofactor_conc
                            flux_values = scan_conc(kinetic_rate_laws, concentrations, reaction_id, a1, conc_samples)

                            # plot M-M curve for this reaction
                            plt.plot(conc_samples, flux_values,
                                 color=colors[index],
                                 label=('%.2e' % (cofactor_conc)))

                        plt.legend(loc='center left', title=cofactor, bbox_to_anchor=(1.15, 0.5), prop={'size': 7})
                        plt.xscale('log')
                        plt.xlabel('%s concentration (M)' % a1)
                        plt.ylabel('flux (M/s)')
                        plt.title('cofactor: %s' % cofactor)

                        col_number += 1

                col_number = 0
                row_number += 1


            if test_transporter:
                concentrations = baseline_concentrations.copy()

                plt.subplot(rows, columns, row_number * columns + col_number + 1)
                for index, transporter_conc in enumerate(transporter_concs_shown):
                    concentrations[transporter] = transporter_conc
                    flux_values = scan_conc(kinetic_rate_laws, concentrations, reaction_id, a1, conc_samples)

                    # plot M-M curve for this reaction
                    plt.plot(conc_samples, flux_values,
                        color = colors[index],
                        label = ('%.2e' % (transporter_conc)))

                plt.legend(loc='center left', title=transporter, bbox_to_anchor=(1.15, 0.5), prop={'size': 7})
                plt.xscale('log')
                plt.xlabel('%s concentration (M)' % a1)
                plt.ylabel('flux (M/s)')
                plt.title('transporter: %s' % transporter)

                col_number += 1


                # Test kcat
                kcat_f = parameters[transporter].get('kcat_f')

                # set concentrations
                concentrations = baseline_concentrations.copy()
                concentrations[transporter] = 1

                cofactor_rel_conc_kms = [1, 2, 100]

                plt.subplot(rows, columns, row_number * columns + col_number + 1)
                for rel_conc in cofactor_rel_conc_kms:

                    for mol_index, mol_id in enumerate(analyze_cofactors):
                        km = parameters[transporter].get(mol_id)
                        concentrations[mol_id] = km * rel_conc  # cofactors set at km * rel_conc

                    for mol_id in analyze_competitors:
                        concentrations[mol_id] = 0  # remove competitors

                    reaction_fluxes = kinetic_rate_laws.get_fluxes(concentrations)
                    flux_value = reaction_fluxes[reaction_id]
                    plt.axhline(y=flux_value, color=np.random.rand(3,), label='[cofactors] = %i*k_m' % (rel_conc))

                plt.axhline(y=kcat_f, color='r', linestyle='dashed', label='k_cat = %.6e' % kcat_f)
                plt.legend(loc='center left', title=transporter, bbox_to_anchor=(1.15, 0.5), prop={'size': 7})
                plt.ylim((0, 2*kcat_f))
                plt.title('[%s] = 1' % transporter)
                plt.tick_params(
                    axis='x',  # changes apply to the x-axis
                    which='both',  # both major and minor ticks are affected
                    bottom=False,  # ticks along the bottom edge are off
                    top=False,  # ticks along the top edge are off
                    labelbottom=False)  # labels along the bottom edge are off

                col_number = 0
                row_number += 1

            if test_competitor:

                for competitor in analyze_competitors:
                    concentrations = baseline_concentrations.copy()

                    plt.subplot(rows, columns, row_number * columns + col_number + 1)
                    for index, competitor_conc in enumerate(conc_shown):
                        concentrations[competitor] = competitor_conc
                        flux_values = scan_conc(kinetic_rate_laws, concentrations, reaction_id, a1, conc_samples)

                        # plot M-M curve for this reaction
                        plt.plot(conc_samples, flux_values,
                            color = colors[index],
                            label = ('%.2e' % (competitor_conc)))

                    plt.legend(loc='center left', title=competitor, bbox_to_anchor=(1.15, 0.5), prop={'size': 7})
                    plt.xscale('log')
                    plt.xlabel(a1 + ' concentration (M)')
                    plt.ylabel('flux (M/s)')
                    plt.title('competitor: %s' % competitor)

                    col_number += 1

                if analyze_competitors:
                    row_number += 1

            col_number = 0

    plt.subplots_adjust(hspace=0.8, wspace=1.1)

    if not os.path.exists(OUTPUT_DIR):
        os.mkdir(OUTPUT_DIR)
    plt.savefig(os.path.join(OUTPUT_DIR, output_filename), bbox_inches='tight')
    plt.savefig(os.path.join(OUTPUT_DIR, output_filename + '.pdf'), bbox_inches='tight')

    print('rate law analysis plot saved')


def scan_conc(rate_law, concentrations, reaction_id, molecule_id, conc_samples):
    # scan the concentrations of molecule_id for a given rate law, and return flux values for the given reaction_id
    flux_values = np.empty_like(conc_samples)
    for idx, conc in enumerate(conc_samples):
        concentrations[molecule_id] = conc
        reaction_fluxes = rate_law.get_fluxes(concentrations)
        flux_values[idx] = reaction_fluxes[reaction_id]

    return flux_values

def load_reactions():
    '''
    Load all reactions, including the locations of enzymes into each reaction's 'catalyzed by' key

    Returns:
        all_reactions(dict)
    '''

    # get protein locations
    proteins_locations = {}
    rows = load_tsv(PROTEINS_FILE)
    for row in rows:
        molecule_id = row["id"]
        location = row["location"]
        molecule_loc = ['{}[{}]'.format(molecule_id, loc) for loc in location]
        proteins_locations[molecule_id] = molecule_loc

    # get complex locations
    rows = load_tsv(COMPLEXATION_FILE)
    for row in rows:
        stoichiometry = row["stoichiometry"]
        for stoich in stoichiometry:
            molecule_id =  stoich['molecule']
            location = stoich["location"]
            molecule_loc = ['{}[{}]'.format(molecule_id, loc) for loc in location]
            proteins_locations[molecule_id] = molecule_loc

    # make dict of all reactions
    all_reactions = {}
    rows = load_tsv(REACTIONS_FILE)
    for row in rows:
        reaction_id = row["reaction id"]
        stoichiometry = row["stoichiometry"]
        reversible = row["is reversible"]
        enzymes = row["catalyzed by"]

        # get location
        enzymes_loc = []
        for enzyme in enzymes:
            if enzyme in proteins_locations.keys():
                enzymes_loc.extend(proteins_locations[enzyme])

        all_reactions[reaction_id] = {
            "stoichiometry": stoichiometry,
            "is reversible": reversible,
            "catalyzed by": enzymes_loc}

    return all_reactions

def get_reactions_from_exchange(all_reactions, include_exchanges):
    '''
    Args:
        all_reactions (dict): all reactions with stoichiometry, reversibility, enzymes
        include_exchanges (list): molecules whose reactions are of interest

    Returns:
        include_reactions (list): all the reactions for molecules listed in include_exchanges

    '''
    include_reactions = []
    for reaction_id, specs in all_reactions.items():
        reaction_molecules = specs['stoichiometry'].keys()
        for exchange in include_exchanges:
            if exchange in reaction_molecules:
                include_reactions.append(reaction_id)
    return include_reactions

def get_molecules_from_reactions(reaction_ids, all_reactions):
    '''
    Inputs:
           reaction_ids (list) -- reaction ids of interest
           all_reactions (dict): all reactions with stoichiometry, reversibility, enzymes
    Returns:
           self.molecule_ids - a list of all molecules used by these reactions
    '''
    molecule_ids = []

    # for reaction_id, specs in all_reactions.items():
    for reaction_id in reaction_ids:
        specs = all_reactions[reaction_id]
        stoichiometry = specs['stoichiometry']
        substrates = stoichiometry.keys()
        enzymes = specs['catalyzed by']
        # Add all relevant molecules_ids
        molecule_ids.extend(substrates)
        molecule_ids.extend(enzymes)
    return list(set(molecule_ids))


class RateLawUtilities(object):

    def __init__(self):

        parser = argparse.ArgumentParser(description='analyze rate laws')
        parser = self.add_arguments(parser)
        args = parser.parse_args()

        # load all reactions
        self.all_reactions = load_reactions()

        # load dict of saved parameters and save parameter file name as param_id
        parameter_file = args.path
        with open(parameter_file, 'r') as fp:
            kinetic_parameters = json.load(fp)
        self.param_id = parameter_file.split(os.path.sep)[-1].split(".")[0]

        # make a dict of reactions that will be configured with the parameters
        make_reaction_ids = kinetic_parameters.keys()
        make_reactions = {
            reaction_id: specs
            for reaction_id, specs in self.all_reactions.items()
            if reaction_id in make_reaction_ids}

        # Make the kinetic model
        self.kinetic_rate_laws = rate_laws.KineticFluxModel(make_reactions, kinetic_parameters)

        # Get list of molecule_ids used by kinetic rate laws
        self.molecule_ids = self.kinetic_rate_laws.molecule_ids

        # make look up object and get saved concentrations from wcm (mmol/L)
        self.look_up = LookUp()
        self.concentrations = self.look_up.look_up('average', args.media, self.molecule_ids)
        for mol_id, conc in self.concentrations.items():
            if conc is None:
                print('{} concentration is not available in look up table, setting to 1 mmol/L'.format(mol_id))
                self.concentrations[mol_id] = 1.0

        if args.analyze:
            self.run_analysis()

        if args.template:
            reactions_list = args.template.split(',')
            self.template_from_reactions(reactions_list)

        if args.exchange_template:
            molecule_list = args.exchange_template.split(',')
            self.template_from_exchange(molecule_list)

    def add_arguments(self, parser):

        parser.add_argument(
            '-a', '--analyze',
            action='store_true',
            default=False,
            help='run analysis on parameter files specified by path')

        parser.add_argument(
            '-t', '--template',
            type=str,
            default='',
            help='A list of reactions for making an empty parameter template, formatted as "reaction_id_1, reaction_id_2"')

        parser.add_argument(
            '-ex', '--exchange_template',
            type=str,
            default='',
            help='A list of exchange molecules for making an empty parameter template, formatted as "molecule_id_1, molecule_id_2"')

        parser.add_argument(
            '-m', '--media',
            type=str,
            default='minimal',
            help='The environment media')

        parser.add_argument(
            '--path',
            type=str,
            default=os.path.join(KINETIC_PARAMETERS_PATH, 'example_parameters.json') ,
            help='the path to the parameter file to be analyzed. parameters available in {}'.format(KINETIC_PARAMETERS_PATH))

        return parser

    def run_analysis(self):

        # run analyses and save output
        analyze_rate_laws(self.kinetic_rate_laws, self.concentrations, self.param_id)

    def template_from_exchange(self, exchange_molecules):
        '''
        saves a rate law parameter template for the list of exchange molecules

        '''

        # get a list of all reactions with exchange_molecules
        reactions_list = get_reactions_from_exchange(self.all_reactions, exchange_molecules)

        self.template_from_reactions(reactions_list)

    def template_from_reactions(self, reactions_list):
        '''
        saves a rate law parameter template for the list of reactions

        '''
        # make a dict of the given reactions using specs from all_reactions
        reactions = {reaction_id: self.all_reactions[reaction_id] for reaction_id in reactions_list}

        # make a parameter template
        parameter_template = self.get_parameter_template(reactions)

        time_stamp = filepath.timestamp()
        output_name = 'parameter_template_' + time_stamp +'.json'
        output_file = os.path.join(OUTPUT_DIR, output_name)
        with open(output_file, 'w') as fp:
            json.dump(parameter_template, fp, sort_keys=True, indent=2)

        print('rate law parameter template saved')

    def get_parameter_template(self, reactions):
        '''
        Given a list of reactions, return a template for required parameters

        Args:
            reactions (dict): a reaction network, with
             {reaction_id: {'catalyzed by': (list), 'is reversible': (bool), 'stoichiometry': (dict)}}

        Returns:
            parameter_template (dict): a template for all parameters required by this rate_law_configuration,
                filled with values of 0.0.

        '''

        rate_law_configuration = rate_laws.make_configuration(reactions)

        parameter_template = {}
        for enzyme_id, configuration in rate_law_configuration.items():
            reaction_cofactors = configuration['reaction_cofactors']
            partition = configuration['partition']

            for reaction_id, cofactors in reaction_cofactors.items():

                # check if reaction is already in the template
                if reaction_id not in parameter_template:
                    parameter_template[reaction_id] = {}

                parameter_template[reaction_id][enzyme_id] = {}
                parameter_template[reaction_id][enzyme_id]['kcat_f'] = None

                reversible = reactions[reaction_id]['is reversible']
                if reversible:
                    parameter_template[reaction_id][enzyme_id]['kcat_r'] = None

                all_bound_molecules = [mol_id for set in partition for mol_id in set]

                for molecule_id in all_bound_molecules:
                    parameter_template[reaction_id][enzyme_id][molecule_id] = None

        return parameter_template


if __name__ == '__main__':
    command = RateLawUtilities()



from __future__ import absolute_import, division, print_function

import os
import sys
import argparse
import uuid

from vivarium.core.composition import (
    agent_environment_experiment,
    simulate_experiment,
    plot_agents_multigen,
    EXPERIMENT_OUT_DIR,
)
from vivarium_cell.plots.multibody_physics import (
    plot_snapshots,
    plot_tags
)

# processes
from vivarium_cell.processes.metabolism import get_minimal_media_iAF1260b

# composites
from vivarium_cell.composites.lattice import (
    Lattice,
    make_lattice_config,
)
from vivarium_cell.composites.growth_division import GrowthDivision
from vivarium_cell.composites.growth_division_minimal import GrowthDivisionMinimal



NAME = 'lattice'
OUT_DIR = os.path.join(EXPERIMENT_OUT_DIR, NAME)
DEFAULT_ENVIRONMENT_TYPE = Lattice
TIME_STEP = 1



# agents and their configurations
agents_library = {
    'growth_division': {
        'name': 'growth_division',
        'type': GrowthDivision,
        'config': {
            'agents_path': ('..', '..', 'agents'),
            'fields_path': ('..', '..', 'fields'),
            'dimensions_path': ('..', '..', 'dimensions'),
        }
    },
    'growth_division_minimal': {
        'name': 'growth_division_minimal',
        'type': GrowthDivisionMinimal,
        'config': {
            'agents_path': ('..', '..', 'agents'),
            'fields_path': ('..', '..', 'fields'),
            'dimensions_path': ('..', '..', 'dimensions'),
            'growth_rate': 0.001,
            'division_volume': 2.6
        }
    },
}


# iAF1260b environment is set to minimal media for the iAF1260b BiGG model
def get_iAF1260b_environment(
    time_step=TIME_STEP,
    bounds=[20, 20],
    n_bins=[10, 10],
    jitter_force=1e-4,
    depth=3000.0,
    scale_concentration=1,  # scales minimal media
    diffusion=5e-3,
    override_initial={},
    keep_fields_emit=[],
):
    media = get_minimal_media_iAF1260b(
        scale_concentration=scale_concentration,
        override_initial=override_initial
    )
    return make_lattice_config(
        time_step=time_step,
        bounds=bounds,
        n_bins=n_bins,
        jitter_force=jitter_force,
        depth=depth,
        diffusion=diffusion,
        concentrations=media,
        keep_fields_emit=keep_fields_emit,
    )

environments_library = {
    'glc_lcts': {
        'type': DEFAULT_ENVIRONMENT_TYPE,
        'config': make_lattice_config(
            molecules=['glc__D_e', 'lcts_e'],
            bounds=[30, 30],
            n_bins=[20, 20],
            jitter_force=1e-5,
        ),
    },
    'iAF1260b': {
        'type': DEFAULT_ENVIRONMENT_TYPE,
        'config': get_iAF1260b_environment(
            bounds=[17, 17],
        ),
    },
    'shallow_iAF1260b': {
        'type': DEFAULT_ENVIRONMENT_TYPE,
        'config': get_iAF1260b_environment(
            time_step=10,
            bounds=[30, 30],
            n_bins=[40, 40],
            jitter_force=2e-3,
            depth=1e1,
            scale_concentration=10000,
            diffusion=1e-1,
            override_initial={
                'glc__D_e': 0.2,
                'lcts_e': 8.0},
            keep_fields_emit=[
                'glc__D_e',
                'lcts_e'],
        ),
    }
}


# simulation settings
def get_experiment_settings(
        experiment_name='lattice',
        description='an experiment in the lattice environment',
        total_time=4000,
        emit_step=10,
        emitter='timeseries',
        agent_names=False,
        return_raw_data=True,
):
    return {
        'experiment_name': experiment_name,
        'description': description,
        'total_time': total_time,
        'emit_step': emit_step,
        'emitter': emitter,
        'agent_names': agent_names,
        'return_raw_data': return_raw_data}


# plot settings
def get_plot_settings(
    skip_paths=[],
    fields=[],
    tags=[],
    n_snapshots=6,
    background_color='black',
):
    settings = {
        'plot_types': {
            'agents': {
                'skip_paths': skip_paths,
                'remove_zeros': True,
            },
        }
    }
    if fields:
        settings['plot_types']['snapshots'] = {
            'fields': fields,
            'n_snapshots': n_snapshots,
        }
    if tags:
        settings['plot_types']['tags'] = {
            'tagged_molecules': tags,
            'n_snapshots': n_snapshots,
            'background_color': background_color,
        }
    return settings

def plot_experiment_output(
        data,
        plot_settings={},
        out_dir='out',
):
    environment_config = plot_settings['environment_config']
    agent_type = plot_settings.get('agent_type', 'agent')
    plot_types = plot_settings['plot_types']

    # extract data
    multibody_config = environment_config['config']['multibody']
    agents = {time: time_data['agents'] for time, time_data in data.items()}
    fields = {time: time_data['fields'] for time, time_data in data.items()}

    # pass to plots
    if 'agents' in plot_types:
        plot_settings = plot_types['agents']
        plot_settings['agents_key'] = 'agents'
        plot_agents_multigen(data, plot_settings, out_dir, agent_type)

    if 'snapshots' in plot_types:
        plot_config = plot_types['snapshots']
        field_ids = plot_types['snapshots']['fields']
        plot_fields = {
            time: {
                field_id: field_instance[field_id]
                for field_id in field_ids}
            for time, field_instance in fields.items()}
        data = {
            'agents': agents,
            'fields': plot_fields,
            'config': multibody_config}
        plot_config.update({
            'out_dir': out_dir,
            'filename': agent_type + '_snapshots',
        })
        plot_snapshots(data, plot_config)

    if 'tags' in plot_types:
        plot_config = plot_types['tags']
        data = {
            'agents': agents,
            'config': multibody_config}
        plot_config.update({
            'out_dir': out_dir,
            'filename': agent_type + '_tags',
        })
        plot_tags(data, plot_config)


# Experiment run function
def run_lattice_experiment(
        agents_config=None,
        environment_config=None,
        initial_state=None,
        initial_agent_state=None,
        experiment_settings=None,
):
    if experiment_settings is None:
        experiment_settings = {}
    if initial_state is None:
        initial_state = {}
    if initial_agent_state is None:
        initial_agent_state = {}

    # agents ids
    agent_ids = []
    for config in agents_config:
        number = config.get('number', 1)
        if 'name' in config:
            name = config['name']
            if number > 1:
                new_agent_ids = [name + '_' + str(num) for num in range(number)]
            else:
                new_agent_ids = [name]
        else:
            new_agent_ids = [str(uuid.uuid1()) for num in range(number)]
        config['ids'] = new_agent_ids
        agent_ids.extend(new_agent_ids)

    # make the experiment
    experiment = agent_environment_experiment(
        agents_config=agents_config,
        environment_config=environment_config,
        initial_state=initial_state,
        initial_agent_state=initial_agent_state,
        settings=experiment_settings,
    )

    # simulate
    return simulate_experiment(
        experiment,
        experiment_settings,
    )


def run_workflow(
        agent_type='growth_division_minimal',
        n_agents=1,
        environment_type='glc_lcts',
        initial_state=None,
        initial_agent_state=None,
        out_dir='out',
        experiment_settings=get_experiment_settings(),
        plot_settings=get_plot_settings()
):
    if initial_state is None:
        initial_state = {}
    if initial_agent_state is None:
        initial_agent_state = {}
    # agent configuration
    agent_config = agents_library[agent_type]
    agent_config['number'] = n_agents
    agents_config = [
        agent_config,
    ]

    # environment configuration
    environment_config = environments_library[environment_type]

    # simulate
    data = run_lattice_experiment(
        agents_config=agents_config,
        environment_config=environment_config,
        initial_state=initial_state,
        initial_agent_state=initial_agent_state,
        experiment_settings=experiment_settings,
    )

    plot_settings['environment_config'] = environment_config
    plot_settings['agent_type'] = agent_type
    plot_experiment_output(
        data,
        plot_settings,
        out_dir,
    )


def test_growth_division_experiment():
    '''test growth_division_minimal agent in lattice experiment'''
    growth_rate = 0.005  # fast!
    total_time = 150

    # get minimal agent config and set growth rate
    agent_config = agents_library['growth_division_minimal']
    agent_config['config']['growth_rate'] = growth_rate
    agent_config['number'] = 1
    agents_config = [agent_config]

    # get environment config
    environment_config = environments_library['glc_lcts']

    # simulate
    experiment_settings = get_experiment_settings(
        total_time=total_time,
        return_raw_data=True)

    data = run_lattice_experiment(
        agents_config=agents_config,
        environment_config=environment_config,
        experiment_settings=experiment_settings)

    # assert division
    time = list(data.keys())
    initial_agents = len(data[time[0]]['agents'])
    final_agents = len(data[time[-1]]['agents'])
    assert final_agents > initial_agents


def make_dir(out_dir):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

def main():
    out_dir = OUT_DIR
    make_dir(out_dir)

    parser = argparse.ArgumentParser(description='lattice_experiment')
    parser.add_argument('--growth_division', '-g', action='store_true', default=False)
    parser.add_argument('--growth_division_minimal', '-m', action='store_true', default=False)
    args = parser.parse_args()
    no_args = (len(sys.argv) == 1)

    if args.growth_division_minimal or no_args:
        minimal_out_dir = os.path.join(out_dir, 'minimal')
        make_dir(minimal_out_dir)
        run_workflow(
            agent_type='growth_division_minimal',
            environment_type='glc_lcts',
            experiment_settings=get_experiment_settings(
                total_time=6000
            ),
            out_dir=minimal_out_dir)

    elif args.growth_division:
        gd_out_dir = os.path.join(out_dir, 'growth_division')
        make_dir(gd_out_dir)
        run_workflow(
            agent_type='growth_division',
            environment_type='glc_lcts',
            experiment_settings=get_experiment_settings(
                total_time=12000
            ),
            plot_settings=get_plot_settings(
                skip_paths=[
                    ('boundary', 'location')
                ],
                fields=[
                    'glc__D_e',
                    'lcts_e',
                ],
                tags=[
                    ('internal', 'protein1'),
                    ('internal', 'protein2'),
                    ('internal', 'protein3'),
                ],
                n_snapshots=5,
            ),
            out_dir=gd_out_dir)



if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

import os

from vivarium.library.units import units
from vivarium.core.process import Generator
from vivarium.core.composition import (
    simulate_compartment_in_experiment,
    plot_agents_multigen,
    COMPARTMENT_OUT_DIR,
)

# processes
from vivarium.processes.meta_division import MetaDivision
from vivarium_cell.processes.growth_protein import GrowthProtein



NAME = 'growth_division_minimal'



class GrowthDivisionMinimal(Generator):

    defaults = {
        # 'growth_rate': 0.006,  # very fast growth
        'boundary_path': ('boundary',),
        'agents_path': ('..', '..', 'agents',),
        'daughter_path': tuple()}

    def __init__(self, config):
        super(GrowthDivisionMinimal, self).__init__(config)

    def generate_processes(self, config):
        # division config
        daughter_path = config['daughter_path']
        agent_id = config['agent_id']
        division_config = dict(
            config.get('division', {}),
            daughter_path=daughter_path,
            agent_id=agent_id,
            compartment=self)

        growth_config = config.get('growth', {})
        if 'growth_rate' in config:
            growth_config['growth_rate'] = config['growth_rate']

        return {
            'growth': GrowthProtein(growth_config),
            'division': MetaDivision(division_config)}

    def generate_topology(self, config):
        boundary_path = config['boundary_path']
        agents_path = config['agents_path']
        return {
            'growth': {
                'internal': ('internal',),
                'global': boundary_path},
            'division': {
                'global': boundary_path,
                'agents': agents_path},
            }


if __name__ == '__main__':
    out_dir = os.path.join(COMPARTMENT_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    agent_id = '0'
    compartment = GrowthDivisionMinimal({'agent_id': agent_id})

    # settings for simulation and plot
    settings = {
        'outer_path': ('agents', agent_id),
        'return_raw_data': True,
        'timestep': 1,
        'total_time': 600}
    output_data = simulate_compartment_in_experiment(compartment, settings)

    plot_settings = {}
    plot_agents_multigen(output_data, plot_settings, out_dir)



from __future__ import absolute_import, division, print_function

import os
import copy
import itertools
import math

import numpy as np
import matplotlib.pyplot as plt

from vivarium.library.units import units
from vivarium.core.composition import simulate_compartment_in_experiment
from vivarium.core.process import Generator

# processes for testing
from vivarium_cell.processes.convenience_kinetics import ConvenienceKinetics, get_glc_lct_config


def get_nested(dict, keys):
    d = dict
    for key in keys[:-1]:
        if key in d:
            d = d[key]
    try:
        value = d[keys[-1]]
    except:
        value = None
        print('value not found for: {}'.format(keys))
    return value

def set_nested(dict, keys, value, create_missing=True):
    d = dict
    for key in keys[:-1]:
        if key in d:
            d = d[key]
        elif create_missing:
            d = d.setdefault(key, {})
        else:
            return dict
    if keys[-1] in d or create_missing:
        d[keys[-1]] = value
    return dict

def get_parameters_logspace(min, max, number):
    ''' get list of n parameters logarithmically spaced between min and max '''
    range = np.logspace(np.log10(min), np.log10(max), number, endpoint=True)
    return list(range)

def run_compartment_get_output(compartment, parameters, condition, metrics, settings):
    settings['initial_state'] = condition
    settings['return_raw_data'] = True
    settings['compartment'] = parameters

    # run the simulation and get the last state
    sim_out = simulate_compartment_in_experiment(compartment, settings)
    time_vec = list(sim_out.keys())
    last_state = sim_out[time_vec[-1]]

    # pull out metric values from last_state
    output = []
    for output_value in metrics:
        output.append(get_nested(last_state, output_value))
    return output


def check_path_exists(path, dict):
    if len(path) > 0:
        key = path[0]
        if key in dict:
            sub_path = path[1:]
            sub_dict = dict[key]
            check_path_exists(sub_path, sub_dict)
        else:
            raise Exception('parameter path does not exist: {}'.format(path))


def parameter_scan(config):
    '''
    Pass in a config (dict) with:
        - compartment (object) -- a compartment class, for configuration by the parameters
        - scan_parameters (dict) -- each parameter location (tuple) mapped to a list of values
        - metrics (list) -- a list of output values (tuple) with the (port, key)
        - conditions (list) -- a list of state values (dict) with {port: {variable: value}}
            for the default state the condition is and empty dict, [{}]
        - settings (dict) -- simulation settings for the experiments

    Returns a list of all parameter combinations, and a dictionary with output values for those parameters
    '''

    compartment = config['compartment']
    scan_params = config['scan_parameters']
    metrics = config['metrics']
    settings = config.get('settings', {})
    conditions = config.get('conditions')
    if not conditions:
        conditions = [{}]
    n_conditions = len(conditions)

    ## Set up the parameters
    # get number of parameter sets for scan
    n_values = [len(v) for v in scan_params.values()]
    n_combinations = np.prod(np.array(n_values))
    print('parameter scan size: {}'.format(n_combinations))

    # get default parameters from baseline compartment object
    default_params = compartment.get_parameters()

    # check that scan parameters are in default parameters
    for path in scan_params.keys():
        check_path_exists(path, default_params)

    # get parameter sets for scan
    param_keys = list(scan_params.keys())
    param_values = list(scan_params.values())
    param_combinations = list(itertools.product(*param_values))  # a list of all parameter combinations
    param_sets = [dict(zip(param_keys, combo)) for combo in param_combinations]  # list of dicts with {param: value}

    # run all parameter sets and save results
    results = []
    for params_index, param_set in enumerate(param_sets):
        # set up the parameters
        # parameters = copy.deepcopy(default_params)
        parameters = {}
        for param_key, param_value in param_set.items():
            parameters = set_nested(parameters, param_key, param_value)

        ## run the parameter set for each condition's state
        for condition_index, condition_state in enumerate(conditions):
            print('running parameter set {}/{}, condition {}/{}'.format(
                params_index + 1,
                n_combinations,
                condition_index+1,
                n_conditions))

            output = run_compartment_get_output(
                compartment,
                parameters,
                condition_state,
                metrics,
                settings)

            result = {
                'parameter_index': params_index,
                'condition_index': condition_index,
                'output': output}
            results.append(result)

    # organize data by metric
    output_data = {
        'results': results,
        'metrics': metrics,
        'parameter_sets': {idx: param_set for idx, param_set in enumerate(param_sets)},
        'conditions': {idx: condition for idx, condition in enumerate(conditions)}}

    return organize_param_scan_results(output_data)


def organize_param_scan_results(data):
    results = data['results']
    metrics = data['metrics']
    parameter_sets = data['parameter_sets']
    conditions = data['conditions']

    param_indices = list(range(0, len(parameter_sets)))
    condition_indices = list(range(0, len(conditions)))

    # organize the results by metric
    metric_data = {
        metric: {
            condition_index: [None for param in param_indices]
            for condition_index in condition_indices}
        for metric in metrics}

    for result in results:
        param_index = result['parameter_index']
        condition_index = result['condition_index']
        output = result['output']
        for metric_index, datum in enumerate(output):
            metric = metrics[metric_index]
            metric_data[metric][condition_index][param_index] = datum

    return {
        'metric_data': metric_data,
        'parameter_sets': parameter_sets,
        'conditions': conditions}


def plot_scan_results(results, out_dir='out', filename='parameter_scan'):
    metric_data = results['metric_data']
    parameter_sets = results['parameter_sets']
    conditions = results['conditions']
    parameter_indices = [idx for idx, param in enumerate(parameter_sets)]

    # make the figure
    n_cols = 1
    lines_per_row = 8
    base_rows = len(metric_data)
    param_rows = math.ceil(len(parameter_indices)/lines_per_row)
    condition_rows = math.ceil(len(conditions)/lines_per_row)
    n_rows = base_rows + param_rows + condition_rows
    fig = plt.figure(figsize=(n_cols * 6, n_rows * 2))
    grid = plt.GridSpec(n_rows, n_cols)
    font = {'size': 6}
    plt.rc('font', **font)

    row_idx = 0
    col_idx = 0
    for metric, data, in metric_data.items():
        ax = fig.add_subplot(grid[row_idx, col_idx])
        for condition, param_data in data.items():
            ax.scatter(parameter_indices, param_data, label=condition)

        ax.legend(title='condition', bbox_to_anchor=(1.2, 1.0))
        ax.set_yscale('log')
        ax.set_ylabel(metric)
        ax.set_xticks(parameter_indices)
        ax.set_xlabel('parameter set #')

        row_idx += 1

    # prepare text
    param_text_row = 1 / lines_per_row / param_rows
    cond_text_row = 1 / lines_per_row / condition_rows
    parameter_text = [
        '{}: {}'.format(param_idx, param_set)
        for param_idx, param_set in parameter_sets.items()]
    condition_text = [
        '{}: {}'.format(condition_idx, condition)
        for condition_idx, condition in conditions.items()]

    # parameter text
    ax = fig.add_subplot(grid[base_rows:base_rows+param_rows, :])
    ax.text(0, 1.0, 'parameters')
    for text_idx, param in enumerate(parameter_text):
        ax.text(0, 0.9-text_idx*param_text_row, param)
    ax.axis('off')

    # condition text
    ax = fig.add_subplot(grid[base_rows+param_rows:, :])
    ax.text(0, 1.0, 'conditions')
    for text_idx, condition in enumerate(condition_text):
        ax.text(0, 0.9-text_idx*cond_text_row, condition)
    ax.axis('off')

    # save the figure
    fig_path = os.path.join(out_dir, filename)
    plt.subplots_adjust(wspace=0.3, hspace=0.5)
    plt.savefig(fig_path, bbox_inches='tight')


# testing
class TestConvienceKinetics(Generator):
    defaults = {
        'boundary_path': ('boundary',),
        'config':  {
            'process': {
                'reactions': {
                    # reaction1 is the reaction ID
                    'reaction1': {
                        'stoichiometry': {
                            # 1 mol A is consumed per mol reaction
                            ('internal', 'A'): -1,
                            ('internal', 'B'): -1,
                            # 2 mol C are produced per mol reaction
                            ('internal', 'C'): 2,
                        },
                        'is reversible': False,
                        'catalyzed by': [
                            ('internal', 'E'),
                        ],
                    }
                },
                'kinetic_parameters': {
                    'reaction1': {
                        ('internal', 'E'): {
                            'kcat_f': 1e1,  # kcat for forward reaction
                            ('internal', 'A'): 1e-1,  # km for A
                            ('internal', 'B'): 1e-1,  # km for B
                        },
                    },
                },
                'initial_state': {
                    'internal': {
                        'A': 1e2,
                        'B': 1e2,
                        'C': 1e2,
                        'E': 1e2,
                    },
                    'fluxes': {
                        'reaction1': 0.0,
                    }
                },
                'ports': {
                    'internal': ['A', 'B', 'C', 'E'],
                    'external': [],
                },
            }
        }
    }

    def __init__(self, config=None):
        self.config = self.defaults['config']
        self.boundary_path = self.or_default(config, 'boundary_path')

    def generate_processes(self, config):
        process = ConvenienceKinetics(config['process'])
        return {'process': process}

    def generate_topology(self, config):
        external_path = self.boundary_path + ('external',)
        return {
            'process': {
                'internal': ('internal',),
                'external': external_path,
                'fields': ('fields',),
                'fluxes': ('fluxes',),
                'global': self.boundary_path,
                'dimensions': ('dimensions',),
            }
        }


def scan_test():
    # initialize the compartment
    compartment = TestConvienceKinetics({})

    # parameters to be scanned, and their values
    scan_params = {
        ('process',
         'kinetic_parameters',
         'reaction1',
         ('internal', 'E'),
         'kcat_f'):
            get_parameters_logspace(1e0, 1e5, 6),
    }

    # metrics are the outputs of a scan
    metrics = [('fluxes', 'reaction1')]

    # define conditions
    conditions = [
        {'internal': {'E': 1e0}},
        {'internal': {'E': 1e1}},
        {'internal': {'E': 1e2}},
    ]

    # set up scan options
    timeline = [(5, {})]
    sim_settings = {
        'timeline': {
            'timeline': timeline,
            'ports': {
                'external': ('boundary', 'external')}}}

    # run scan
    scan_config = {
        'compartment': compartment,
        'scan_parameters': scan_params,
        'conditions': conditions,
        'metrics': metrics,
        'settings': sim_settings}
    results = parameter_scan(scan_config)
    return results


if __name__ == '__main__':
    out_dir = os.path.join('out', 'parameters', 'scan')
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    results = scan_test()
    plot_scan_results(results, out_dir, 'test_scan')



"""
==========================
Multibody physics process
==========================
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import argparse

import random
import math

import numpy as np

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches

# vivarium imports
from vivarium.library.units import units, remove_units
from vivarium.core.emitter import timeseries_from_data
from vivarium.core.process import Process
from vivarium.core.composition import (
    process_in_experiment,
    simulate_experiment,
    PROCESS_OUT_DIR,
)

# vivarium-cell imports
from vivarium_cell.processes.derive_globals import volume_from_length
from vivarium_cell.library.pymunk_multibody import PymunkMultibody
from vivarium_cell.plots.multibody_physics import (
    check_plt_backend,
    plot_agent,
    plot_agents,
    plot_snapshots,
    plot_temporal_trajectory,
)


NAME = 'multibody'

DEFAULT_BOUNDS = [10, 10]

# constants
PI = math.pi



def random_body_position(body):
    # pick a random point along the boundary
    width, length = body.dimensions
    if random.randint(0, 1) == 0:
        # force along ends
        if random.randint(0, 1) == 0:
            # force on the left end
            location = (random.uniform(0, width), 0)
        else:
            # force on the right end
            location = (random.uniform(0, width), length)
    else:
        # force along length
        if random.randint(0, 1) == 0:
            # force on the bottom end
            location = (0, random.uniform(0, length))
        else:
            # force on the top end
            location = (width, random.uniform(0, length))
    return location


def daughter_locations(parent_location, parent_values):
    parent_length = parent_values['length']
    parent_angle = parent_values['angle']
    pos_ratios = [-0.25, 0.25]
    daughter_locations = []
    for daughter in range(2):
        dx = parent_length * pos_ratios[daughter] * math.cos(parent_angle)
        dy = parent_length * pos_ratios[daughter] * math.sin(parent_angle)
        location = [parent_location[0] + dx, parent_location[1] + dy]
        daughter_locations.append(location)
    return daughter_locations



class Multibody(Process):
    """Simulates collisions and forces between agent bodies with a multi-body physics engine.

    :term:`Ports`:
    * ``agents``: The store containing all agent sub-compartments. Each agent in
      this store has values for location, angle, length, width, mass, thrust, and torque.

    Arguments:
        initial_parameters(dict): Accepts the following configuration keys:

        * **jitter_force**: force applied to random positions along agent
          bodies to mimic thermal fluctuations. Produces Brownian motion.
        * **agent_shape** (:py:class:`str`): agents can take the shapes
          ``rectangle``, ``segment``, or ``circle``.
        * **bounds** (:py:class:`list`): size of the environment in
          micrometers, with ``[x, y]``.
        * **mother_machine** (:py:class:`bool`): if set to ``True``, mother
          machine barriers are introduced.
        * ***animate*** (:py:class:`bool`): interactive matplotlib option to
          animate multibody. To run with animation turned on set True, and use
          the TKAgg matplotlib backend:

          .. code-block:: console

              $ MPLBACKEND=TKAgg python vivarium/processes/multibody_physics.py

    Notes:
        * rotational diffusion in liquid medium with viscosity = 1 mPa.s: :math:`Dr = 3.5 \pm0.3 rad^{2}/s`
          (Saragosti, et al. 2012. Modeling E. coli tumbles by rotational diffusion.)
        * translational diffusion in liquid medium with viscosity = 1 mPa.s: :math:`Dt = 100 um^{2}/s`
          (Saragosti, et al. 2012. Modeling E. coli tumbles by rotational diffusion.)
    """

    name = NAME
    defaults = {
        'jitter_force': 1e-3,  # pN
        'agent_shape': 'segment',
        'bounds': DEFAULT_BOUNDS,
        'mother_machine': False,
        'animate': False,
        'time_step': 2,
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        # multibody parameters
        jitter_force = self.or_default(
            initial_parameters, 'jitter_force')
        self.agent_shape = self.or_default(
            initial_parameters, 'agent_shape')
        self.bounds = self.or_default(
            initial_parameters, 'bounds')
        self.mother_machine = self.or_default(
            initial_parameters, 'mother_machine')

        # make the multibody object
        self.time_step = self.or_default(
            initial_parameters, 'time_step')
        multibody_config = {
            'agent_shape': self.agent_shape,
            'jitter_force': jitter_force,
            'bounds': self.bounds,
            'barriers': self.mother_machine,
            'physics_dt': self.time_step / 10,
        }
        self.physics = PymunkMultibody(multibody_config)

        # interactive plot for visualization
        self.animate = initial_parameters.get('animate', self.defaults['animate'])
        if self.animate:
            plt.ion()
            self.ax = plt.gca()
            self.ax.set_aspect('equal')

        parameters = {'time_step': self.defaults['time_step']}
        parameters.update(initial_parameters)

        super(Multibody, self).__init__(parameters)

    def ports_schema(self):
        glob_schema = {
            '*': {
                'boundary': {
                    'location': {
                        '_emit': True,
                        '_default': [0.5 * bound for bound in self.bounds],
                        '_updater': 'set',
                        '_divider': {
                            'divider': daughter_locations,
                            'topology': {
                                'length': ('length',),
                                'angle': ('angle',)}}},
                    'length': {
                        '_emit': True,
                        '_default': 2.0,
                        '_divider': 'split',  # TODO -- might want this to be set by agent
                        '_updater': 'set'},
                    'width': {
                        '_emit': True,
                        '_default': 1.0,
                        '_updater': 'set'},
                    'angle': {
                        '_emit': True,
                        '_default': 0.0,
                        '_updater': 'set'},
                    'mass': {
                        '_emit': True,
                        '_default': 1339 * units.fg,
                        '_updater': 'set'},
                    'thrust': {
                        '_default': 0.0,
                        '_updater': 'set'},
                    'torque': {
                        '_default': 0.0,
                        '_updater': 'set'},
                }
            }
        }
        schema = {'agents': glob_schema}

        return schema

    def next_update(self, timestep, states):
        agents = states['agents']

        # animate before update
        if self.animate:
            self.animate_frame(agents)

        # update multibody with new agents
        self.physics.update_bodies(remove_units(agents))

        # run simulation
        self.physics.run(timestep)

        # get new agent positions
        agent_positions = self.physics.get_body_positions()
        update = {'agents': agent_positions}

        # for mother machine configurations, remove cells above the channel height
        if self.mother_machine:
            channel_height = self.mother_machine['channel_height']
            delete_agents = []
            for agent_id, position in agent_positions.items():
                location = position['boundary']['location']
                y_loc = location[1]
                if y_loc > channel_height:
                    # cell has moved past the channels
                    delete_agents.append(agent_id)
            if delete_agents:
                update['agents'] = {
                    agent_id: position
                    for agent_id, position in agent_positions.items()
                    if agent_id not in delete_agents}

                update['agents']['_delete'] = [
                    (agent_id,)
                    for agent_id in delete_agents]

        return update

    ## matplotlib interactive plot
    def animate_frame(self, agents):
        plt.cla()
        for agent_id, data in agents.items():
            # location, orientation, length
            data = data['boundary']
            x_center = data['location'][0]
            y_center = data['location'][1]
            angle = data['angle'] / PI * 180 + 90  # rotate 90 degrees to match field
            length = data['length']
            width = data['width']

            # get bottom left position
            x_offset = (width / 2)
            y_offset = (length / 2)
            theta_rad = math.radians(angle)
            dx = x_offset * math.cos(theta_rad) - y_offset * math.sin(theta_rad)
            dy = x_offset * math.sin(theta_rad) + y_offset * math.cos(theta_rad)

            x = x_center - dx
            y = y_center - dy

            if self.agent_shape == 'rectangle' or self.agent_shape == 'segment':
                # Create a rectangle
                rect = patches.Rectangle((x, y), width, length, angle=angle, linewidth=1, edgecolor='b')
                self.ax.add_patch(rect)

            elif self.agent_shape == 'circle':
                # Create a circle
                circle = patches.Circle((x, y), width, linewidth=1, edgecolor='b')
                self.ax.add_patch(circle)

        plt.xlim([0, self.bounds[0]])
        plt.ylim([0, self.bounds[1]])
        plt.draw()
        plt.pause(0.01)


# configs
def make_random_position(bounds):
    return [
        np.random.uniform(0, bounds[0]),
        np.random.uniform(0, bounds[1])]

def single_agent_config(config):
    # cell dimensions
    width = 1
    length = 2
    volume = volume_from_length(length, width)
    bounds = config.get('bounds', DEFAULT_BOUNDS)
    location = config.get('location')
    if location:
        location = [loc * bounds[n] for n, loc in enumerate(location)]
    else:
        location = make_random_position(bounds)

    return {'boundary': {
        'location': location,
        'angle': np.random.uniform(0, 2 * PI),
        'volume': volume,
        'length': length,
        'width': width,
        'mass': 1339 * units.fg,
        'thrust': 0,
        'torque': 0}}

def agent_body_config(config):
    agent_ids = config['agent_ids']
    agent_config = {
        agent_id: single_agent_config(config)
        for agent_id in agent_ids}
    return agent_config

def get_baseline_config(config={}):
    animate = config.get('animate', False)
    bounds = config.get('bounds', [500, 500])
    jitter_force = config.get('jitter_force', 0)
    n_agents = config.get('n_agents', 1)
    initial_location = config.get('initial_location')

    # agent settings
    agent_ids = [str(agent_id) for agent_id in range(n_agents)]
    motility_config = {
        'animate': animate,
        'jitter_force': jitter_force,
        'bounds': bounds}
    body_config = {
        'bounds': bounds,
        'agent_ids': agent_ids,
        'location': initial_location}
    motility_config['agents'] = agent_body_config(body_config)
    return motility_config

# tests and simulations
def test_multibody(config={'n_agents':1}, time=10):
    n_agents = config.get('n_agents',1)
    agent_ids = [str(agent_id) for agent_id in range(n_agents)]

    body_config = {
        'agents': agent_body_config({'agent_ids': agent_ids})}
    multibody = Multibody(body_config)

    # initialize agent's boundary state
    initial_agents_state = body_config['agents']
    experiment = process_in_experiment(multibody)
    experiment.state.set_value({'agents': initial_agents_state})

    # run experiment
    settings = {
        'timestep': 1,
        'total_time': time,
        'return_raw_data': True}
    return simulate_experiment(experiment, settings)

def simulate_growth_division(config, settings):
    initial_agents_state = config['agents']

    # make the process
    multibody = Multibody(config)
    experiment = process_in_experiment(multibody)
    experiment.state.update_subschema(
        ('agents',), {
            'boundary': {
                'mass': {
                    '_divider': 'split'},
                'length': {
                    '_divider': 'split'}}})
    experiment.state.apply_subschemas()
    process_topology = {'agents': ('agents',)}  # TODO -- get topology??

    # get initial agent state
    experiment.state.set_value({'agents': initial_agents_state})
    agents_store = experiment.state.get_path(['agents'])

    ## run simulation
    # get simulation settings
    growth_rate = settings.get('growth_rate', 0.0006)
    growth_rate_noise = settings.get('growth_rate_noise', 0.0)
    division_volume = settings.get('division_volume', 0.4)
    total_time = settings.get('total_time', 10)
    timestep = 1

    time = 0
    while time < total_time:
        experiment.update(timestep)
        time += timestep
        agents_state = agents_store.get_value()

        agent_updates = {}
        remove_agents = []
        add_agents = {}
        for agent_id, state in agents_state.items():
            state = state['boundary']
            location = state['location']
            angle = state['angle']
            length = state['length']
            width = state['width']
            mass = state['mass'].magnitude

            # update
            growth_rate2 = (growth_rate + np.random.normal(0.0, growth_rate_noise)) * timestep
            new_mass = mass + mass * growth_rate2
            new_length = length + length * growth_rate2
            new_volume = volume_from_length(new_length, width)

            if new_volume > division_volume:
                daughter_ids = [str(agent_id) + '0', str(agent_id) + '1']

                daughter_updates = []
                for daughter_id in daughter_ids:
                    daughter_updates.append({
                        'daughter': daughter_id,
                        'path': (daughter_id,),
                        'processes': {},
                        'topology': {},
                        'initial_state': {}})

                # initial state will be provided by division in the tree
                update = {
                    '_divide': {
                        'mother': agent_id,
                        'daughters': daughter_updates}}
                invoked_update = InvokeUpdate({'agents': update})
                update_tuples = [(invoked_update, process_topology, agents_store)]
                experiment.send_updates(update_tuples)
            else:
                agent_updates[agent_id] = {
                    'boundary': {
                        'volume': new_volume,
                        'length': new_length,
                        'mass': new_mass * units.fg}}

        # update experiment
        invoked_update = InvokeUpdate({'agents': agent_updates})
        update_tuples = [(invoked_update, process_topology, agents_store)]
        experiment.send_updates(update_tuples)

    return experiment.emitter.get_data()


class InvokeUpdate(object):
    def __init__(self, update):
        self.update = update
    def get(self, timeout=0):
        return self.update

def tumble(tumble_jitter=120.0):
    thrust = 100  # pN
    torque = random.normalvariate(0, tumble_jitter)
    return [thrust, torque]

def run():
    # average thrust = 200 pN according to:
    # Berg, Howard C. E. coli in Motion. Under "Torque-Speed Dependence"
    thrust = 250  # pN
    torque = 0.0
    return [thrust, torque]

def simulate_motility(config, settings):
    # time of motor behavior without chemotaxis
    run_time = 0.42  # s (Berg)
    tumble_time = 0.14  # s (Berg)

    total_time = settings['total_time']
    timestep = settings['timestep']
    initial_agents_state = config['agents']

    # make the process
    config['time_step'] = timestep
    multibody = Multibody(config)
    experiment = process_in_experiment(multibody)
    experiment.state.update_subschema(
        ('agents',), {
            'boundary': {
                'thrust': {
                    '_emit': True,
                    '_updater': 'set'},
                'torque': {
                    '_emit': True,
                    '_updater': 'set'}},
            'cell': {
                'motor_state': {
                    '_value': 0,
                    '_updater': 'set',
                    '_emit': True,
                }}})
    experiment.state.apply_subschemas()
    process_topology = {'agents': ('agents',)}  # TODO -- get topology??

    # get initial agent state
    experiment.state.set_value({'agents': initial_agents_state})
    agents_store = experiment.state.get_path(['agents'])

    # initialize hidden agent motile states, and update agent motile forces in agent store
    agent_motile_states = {}
    motile_forces = {}
    for agent_id, specs in agents_store.get_value().items():
        [thrust, torque] = run()
        agent_motile_states[agent_id] = {
            'motor_state': 1,  # 0 for run, 1 for tumble
            'time_in_motor_state': 0}
        motile_forces[agent_id] = {
            'boundary': {
                'thrust': thrust,
                'torque': torque},
            'cell': {
                'motor_state': 1}}

    invoked_update = InvokeUpdate({'agents': motile_forces})
    update_tuples = [(invoked_update, process_topology, agents_store)]
    experiment.send_updates(update_tuples)

    ## run simulation
    # test run/tumble
    time = 0
    while time < total_time:
        experiment.update(timestep)
        time += timestep

        # update motile force and apply to state
        motile_forces = {}
        for agent_id, motile_state in agent_motile_states.items():
            motor_state = motile_state['motor_state']
            time_in_motor_state = motile_state['time_in_motor_state']
            thrust = None

            if motor_state == 1:  # tumble
                if time_in_motor_state < tumble_time:
                    # [thrust, torque] = tumble()
                    time_in_motor_state += timestep
                else:
                    # switch
                    [thrust, torque] = run()
                    motor_state = 0
                    time_in_motor_state = 0

            elif motor_state == 0:  # run
                if time_in_motor_state < run_time:
                    # [thrust, torque] = run()
                    time_in_motor_state += timestep
                else:
                    # switch
                    [thrust, torque] = tumble()
                    motor_state = 1
                    time_in_motor_state = 0

            agent_motile_states[agent_id] = {
                'motor_state': motor_state,  # 0 for run, 1 for tumble
                'time_in_motor_state': time_in_motor_state}
            motile_forces[agent_id] = {
                'cell': {
                    'motor_state': motor_state}}

            if thrust:
                motile_forces[agent_id]['boundary'] = {
                    'thrust': thrust,
                    'torque': torque}

        invoked_update = InvokeUpdate({'agents': motile_forces})
        update_tuples = [(invoked_update, process_topology, agents_store)]
        experiment.send_updates(update_tuples)

    return experiment.emitter.get_data()

def run_jitter(config={}, out_dir='out', filename='jitter'):
    total_time = config.get('total_time', 30)
    timestep = config.get('timestep', 0.05)
    motility_config = get_baseline_config({
        'animate': False,
        'jitter_force': 1e0,
        'bounds': [50, 50],
        'n_agents': 8,
    })

    # make the process
    multibody = Multibody(motility_config)
    experiment = process_in_experiment(multibody)
    experiment.state.update_subschema(
        ('agents',), {
            'cell': {
                'motor_state': {
                    '_value': 0,
                    '_updater': 'set',
                    '_emit': True,
                }}})
    experiment.state.apply_subschemas()

    time = 0
    while time < total_time:
        experiment.update(timestep)
        time += timestep
    data = experiment.emitter.get_data()

    # make trajectory plot
    timeseries = timeseries_from_data(data)
    plot_temporal_trajectory(timeseries, motility_config, out_dir, filename + '_trajectory')

def run_motility(config={}, out_dir='out', filename='motility'):
    total_time = config.get('total_time', 30)
    timestep = config.get('timestep', 0.05)
    config['initial_location'] = [0.5, 0.5]
    motility_config = get_baseline_config(config)

    # simulation settings
    motility_sim_settings = {
        'timestep': timestep,
        'total_time': total_time}

    # run motility sim
    motility_data = simulate_motility(motility_config, motility_sim_settings)
    motility_timeseries = timeseries_from_data(motility_data)

    plot_temporal_trajectory(motility_timeseries, motility_config, out_dir, filename + '_trajectory')

def run_growth_division(config={}):
    n_agents = 1
    agent_ids = [str(agent_id) for agent_id in range(n_agents)]

    bounds = [20, 20]
    settings = {
        'growth_rate': 0.02,
        'growth_rate_noise': 0.02,
        'division_volume': 2.6,
        'total_time': 140}

    gd_config = {
        'agent_shape': config.get('agent_shape', 'segment'),
        'animate': True,
        'jitter_force': 1e-3,
        'bounds': bounds}
    body_config = {
        'bounds': bounds,
        'agent_ids': agent_ids}
    gd_config['agents'] = agent_body_config(body_config)
    gd_data = simulate_growth_division(gd_config, settings)

    # snapshots plot
    agents = {
        time: time_data['agents']
        for time, time_data in gd_data.items()
        if bool(time_data)}
    data = {
        'agents': agents,
        'config': gd_config}
    plot_config = {
        'out_dir': out_dir,
        'filename': 'growth_division_snapshots'}
    plot_snapshots(data, plot_config)

if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    parser = argparse.ArgumentParser(description='multibody')
    parser.add_argument('--circles', '-c', action='store_true', default=False)
    parser.add_argument('--motility', '-m', action='store_true', default=False)
    parser.add_argument('--growth', '-g', action='store_true', default=False)
    parser.add_argument('--scales', '-s', action='store_true', default=False)
    parser.add_argument('--jitter', '-j', action='store_true', default=False)
    args = parser.parse_args()
    no_args = (len(sys.argv) == 1)

    if args.motility or no_args:
        run_motility({'animate': True}, out_dir)
    if args.growth or no_args:
        run_growth_division()
    if args.circles:
        run_growth_division({'agent_shape': 'circle'})
    if args.jitter:
        run_jitter({}, out_dir, 'jitter')
    if args.scales:
        bounds = [1000, 1000]
        jitter_force = 0

        ts_0p1 = {
            'timestep': 0.1,
            'bounds': bounds,
            'jitter_force': jitter_force}
        run_motility(ts_0p1, out_dir, 'ts_0p1')

        ts_0p01 = {
            'timestep': 0.01,
            'bounds': bounds,
            'jitter_force': jitter_force}
        run_motility(ts_0p01, out_dir, 'ts_0p01')

        # bounds_500 = {'bounds': [500, 500]}
        # run_motility(bounds_500, out_dir, 'bounds_500')
        #
        # bounds_5000 = {'bounds': [5000, 5000]}
        # run_motility(bounds_5000, out_dir, 'bounds_5000')



import os
import csv

from vivarium_cell.data.spreadsheets import load_tsv

FLAT_DIR = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data", "flat")

LIST_OF_FLAT_FILENAMES = (
    os.path.join("wcEcoli_genes.tsv"),
    os.path.join("wcEcoli_proteins.tsv"),
    os.path.join("wcEcoli_environment_molecules.tsv"),
    os.path.join("timelines_def.tsv"),
    os.path.join("media_recipes.tsv"),
    os.path.join("media", "wcEcoli_base.tsv"),
    os.path.join("media", "M9.tsv"),
    os.path.join("media", "M9_GLC.tsv"),
    os.path.join("media", "5X_supplement_EZ.tsv"),
    os.path.join("media", "GLC_G6P.tsv"),
    os.path.join("media", "GLC_LCT.tsv"),
    os.path.join("media", "ecoli_core_GLC.tsv"),
    os.path.join("media", "PURE_Fuji_2014.tsv"),
    os.path.join("media", "PURE_Ueda_2010.tsv"),
)

class DataStore(object):
    def __init__(self):
        pass

class KnowledgeBase(object):
    """ KnowledgeBase """

    def __init__(self):
        # Load raw data from TSV files

        for filename in LIST_OF_FLAT_FILENAMES:
            self._load_tsv(FLAT_DIR, filename)

        self.genes = {
            gene['symbol']: gene
            for gene in self.wcEcoli_genes}

        self.proteins = {
            protein['geneId']: protein
            for protein in self.wcEcoli_proteins}

    def _load_tsv(self, dir_name, file_name):
        path = self
        steps = file_name.split(os.path.sep)
        for subPath in steps[:-1]:
            if not hasattr(path, subPath):
                setattr(path, subPath, DataStore())
            path = getattr(path, subPath)
        attrName = steps[-1].split(".")[0]
        setattr(path, attrName, [])

        file_path = os.path.join(dir_name, file_name)
        rows = load_tsv(file_path)
        setattr(path, attrName, [row for row in rows])

    def concatenate_sequences(self, units):
        sequence = ''
        for unit in units:
            gene = self.genes[unit]
            protein = self.proteins[gene['id']]
            sequence += protein['seq']
        return sequence




from __future__ import absolute_import, division, print_function

import os

import numpy as np

from vivarium.library.units import units
from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    plot_simulation_output,
    PROCESS_OUT_DIR,
)
from vivarium_cell.processes.derive_globals import AVOGADRO


NAME = 'growth_protein'


class GrowthProtein(Process):

    name = NAME
    defaults = {
        'initial_mass': 1339 * units.fg,
        # 'initial_protein': 3.9e7,  # counts of protein
        # the median E. coli protein is 209 amino acids long, and AAs ~ 100 Da
        'protein_mw': 2.09e4 * units.g / units.mol,
        'growth_rate': 0.000275,  # for doubling time about every 2520 seconds
        'global_deriver_key': 'global_deriver',
        'mass_deriver_key': 'mass_deriver',
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        self.growth_rate = self.or_default(initial_parameters, 'growth_rate')
        self.global_deriver_key = self.or_default(
            initial_parameters, 'global_deriver_key')
        self.mass_deriver_key = self.or_default(
            initial_parameters, 'mass_deriver_key')

        # default state
        # 1000 proteins per fg
        initial_mass = self.or_default(
            initial_parameters, 'initial_mass')
        self.protein_mw = self.or_default(
            initial_parameters, 'protein_mw')
        self.initial_protein = (initial_mass.to('g') / self.protein_mw * AVOGADRO)
        self.divide_protein = self.initial_protein * 2

        parameters = {
            'growth_rate': self.growth_rate}
        parameters.update(initial_parameters)

        super(GrowthProtein, self).__init__(parameters)

    def ports_schema(self):
        return {
            'internal': {
                'protein': {
                    '_default': self.initial_protein,
                    '_divider': 'split',
                    '_emit': True,
                    '_properties': {
                        'mw': self.protein_mw}}},
            'global': {
                'volume': {
                    '_updater': 'set',
                    '_divider': 'split'},
                'divide': {
                    '_default': False,
                    '_updater': 'set'}}}

    def derivers(self):
        return {
            self.mass_deriver_key: {
                'deriver': 'mass_deriver',
                'port_mapping': {
                    'global': 'global'},
                'config': {}},
            self.global_deriver_key: {
                'deriver': 'globals_deriver',
                'port_mapping': {
                    'global': 'global'},
                'config': {}},
        }

    def next_update(self, timestep, states):
        protein = states['internal']['protein']
        total_protein = protein * np.exp(self.parameters['growth_rate'] * timestep)
        new_protein = int(total_protein - protein)
        extra = total_protein - int(total_protein)

        # simulate remainder
        where = np.random.random()
        if where < extra:
            new_protein += 1

        divide = False
        if protein >= self.divide_protein:
            divide = True

        return {
            'internal': {
                'protein': new_protein},
            'global': {
                'divide': divide}}

if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    process = GrowthProtein()
    settings = {'total_time': 2520}
    timeseries = simulate_process_in_experiment(process, settings)

    volume_ts = timeseries['global']['volume']
    mass_ts = timeseries['global']['mass']
    print('volume growth: {}'.format(volume_ts[-1] / volume_ts[0]))
    print('mass growth: {}'.format(mass_ts[-1] / mass_ts[0]))

    plot_simulation_output(timeseries, {}, out_dir)
